#!/usr/bin/env python3

import json
import re
import sys
from enum import Enum
from math import floor, log10
from typing import Dict, List

"""
Read "smartctl" output:
"""


class PORT(Enum):
    SATA = "sata-ports-n"
    MSATA = "msata-ports-n"
    IDE = "ide-ports-n"
    MINIIDE = "mini-ide-ports-n"
    M2 = "m2-connectors-n"
    # TODO: add more, if they can even be detected


def parse_smartctl(file: str, interactive: bool = False) -> List[dict]:
    """
    Parse a list of smartctl outputs to a list of disks.
    """
    disks = []
    jdisks = json.loads(file)
    for jdisk in jdisks:
        disk = parse_single_disk(jdisk, interactive)
        disks.append(disk)
    return disks


def seagate_model_decode(disk, model: str, maxtor: bool = False):
    # TODO: SCSI formats (https://hardforum.com/threads/understanding-hard-drive-model-numbers.921544/)
    old_format = r"STM?([0-9])[0-9]{2,}([A-Z][A-Z]?)*.*"

    if re.match(old_format, model):
        result = re.search(f"^{old_format}", model)
        ff_num = result.group(1)
        interface_num = result.group(2)

        ff = {
            "3": "3.5",
            "6": "1",
            "7": "1.8",
            "9": "2.5",
        }.get(ff_num)
        _add_feature_if_possible(disk, "hdd-form-factor", ff)

        if interface_num == "AS":
            _add_interface_if_possible(disk, "sata-ports-n")
        elif interface_num == "A":
            _add_interface_if_possible(disk, "ide-ports-n")

    if not maxtor:
        rpm = re.search(r"(?:.+ )*([1-9][0-9]00)\.[0-9]", disk.get("family", ""))
        if rpm:
            _add_feature_if_possible(disk, "spin-rate-rpm", int(rpm.group(1)))


def samsung_model_decode(disk, model: str):
    # SV + capacity (3) + number of heads (platters*2) + interface?
    sv_format = r"SV[0-9]{2,4}([A-Z])"

    if re.match(sv_format, model):
        result = re.search(f"^{sv_format}", model)

        if result.group(1) in ("C",):
            _add_interface_if_possible(disk, "sata-ports-n")
        elif result.group(1) in ("D", "E", "N", "H"):
            _add_interface_if_possible(disk, "ide-ports-n")


def hitachi_model_decode(disk, model: str):
    # https://www.instantfundas.com/2009/02/how-to-interpret-hard-disk-model.html
    h_format = r"H[A-Z][A-Z]([0-9]{2})[0-9]{2}[0-9]{2}[A-Z]([A-Z0-9])([A-Z][A-Z0-9])[0-9]+"
    ic25_format = r"IC(25|35)([A-Z])[0-9]{3}([A-Z0-9]{2})[A-Z0-9]{2}([0-9]{2})-?.*"

    if re.match(h_format, model):
        result = re.search(f"^{h_format}", model)
        rpm_num = result.group(1)
        ff_num = result.group(2)
        interface_num = result.group(3)

        rpm = {
            "36": 3600,
            "42": 4200,
            "54": 5400,
            "72": 7200,
            "10": 10000,
            "15": 15000,
        }.get(rpm_num)
        _add_feature_if_possible(disk, "spin-rate-rpm", rpm)

        ff = {
            "L": ("3.5", None),
            "S": ("2.5", 15),
            "9": ("2.5", 9.5),
            "8": ("2.5", 8),
            "7": ("2.5", 7),
            "5": ("2.5", 5),
        }.get(ff_num)
        if ff:
            _add_feature_if_possible(disk, "hdd-form-factor", ff[0])
            _add_feature_if_possible(disk, "height-mm", ff[1])

        if interface_num in ("A3", "SA"):
            _add_interface_if_possible(disk, "sata-ports-n")
        elif interface_num == "AT":
            _add_interface_if_possible(disk, "ide-ports-n")
    elif re.match(ic25_format, model):
        result = re.search(f"^{ic25_format}", model)
        ff_num = result.group(1)
        h_num = result.group(2)
        interface_num = result.group(3)
        rpm_num = result.group(4)

        ff = {
            "25": "2.5",
            "35": "3.5",
        }.get(ff_num)
        _add_feature_if_possible(disk, "hdd-form-factor", ff)

        h = {
            "L": 25.4,
            "T": 12.5,
            "N": 9.5,
        }.get(h_num)
        _add_feature_if_possible(disk, "height-mm", h)

        rpm = {
            "04": 4200,
            "05": 5400,
        }.get(rpm_num)
        _add_feature_if_possible(disk, "spin-rate-rpm", rpm)

        if interface_num in ("AV", "AT"):
            if ff == "2.5":
                _add_interface_if_possible(disk, "mini-ide-ports-n")
            else:
                _add_interface_if_possible(disk, "ide-ports-n")
        elif interface_num == "UC":
            _add_interface_if_possible(disk, "scsi-sca2-ports-n")
        elif interface_num == "UW":
            _add_interface_if_possible(disk, "scsi-db68-ports-n")


def toshiba_model_decode(disk, model: str):
    mk_format = r"MK[0-9]{2}[0-9]{2}G([A-Z])([A-Z])[A-Z]?"

    if re.match(mk_format, model):
        result = re.search(f"^{mk_format}", model)
        interface_num = result.group(1)
        ff_num = result.group(2)

        ff = {
            "A": ("1.8", 5, 3600),
            "B": ("1.8", 8, 3600),
            "G": ("1.8", 8, 5400),
            "H": ("1.8", 8, 4200),
            "L": ("1.8", 5, 4200),
            "K": ("3.5", None, 7200),
            "M": ("2.5", 12.7, 5400),
            "P": ("2.5", None, 4200),
            "R": ("2.5", None, 15000),
            "S": ("2.5", 9.5, 4200),
            "X": ("2.5", 9.5, 5400),
            "Y": ("2.5", 9.5, 7200),
        }.get(ff_num)
        if ff:
            _add_feature_if_possible(disk, "hdd-form-factor", ff[0])
            _add_feature_if_possible(disk, "height-mm", ff[1])
            _add_feature_if_possible(disk, "spin-rate-rpm", ff[2])

        interface = {
            "A": "mini-ide-ports-n" if ff and ff[0] != "3.5" else "ide-ports-n",
            "P": "mini-ide-ports-n",
            "R": "sas-sata-ports-n",
            "S": "sata-ports-n",
        }.get(interface_num)
        _add_interface_if_possible(disk, interface)


def fujitsu_model_decode(disk, model: str):
    mhx2_format = r"MH[A-Z](2|3)[0-9]{3}([A-Z]{2})U?(?: .+)?"

    if re.match(mhx2_format, model):
        result = re.search(f"^{mhx2_format}", model)
        ff_num = result.group(1)
        extra_num = result.group(2)

        ff = {
            "2": "2.5",
            "3": "3.5",
        }.get(ff_num)
        _add_feature_if_possible(disk, "hdd-form-factor", ff)

        ide = "ide-ports-n"
        if ff == "2.5":
            ide = "mini-ide-ports-n"

        extra = {
            "AH": (ide, 5400),
            "AS": (ide, 5400),
            "AT": (ide, 4200),
            "BH": ("sata-ports-n", 5400),
            "BS": ("sata-ports-n", 5400),
            "BT": ("sata-ports-n", 4200),
        }.get(extra_num)
        if extra:
            _add_interface_if_possible(disk, extra[0])
            _add_feature_if_possible(disk, "spin-rate-rpm", extra[1])

        # _add_feature_if_possible(disk, "height-mm", 9.5)


def quantum_model_decode(disk, model: str):
    if model.lower().startswith("fireball"):
        _add_interface_if_possible(disk, "ide-ports-n")


def wd_model_decode(disk, model: str):
    old_format = r"WD[0-9]{2,4}([A-Z])([A-Z])-.+"
    new_format = r"WD[0-9]{4}([A-Z])[A-Z]([A-Z])([A-Z])-.+"

    if re.match(old_format, model):
        result = re.search(f"^{old_format}", model)
        rpm_num = result.group(1)
        interface_num = result.group(2)

        rpm = {
            "A": 5400,
            "B": 7200,
            "C": 10000,
            "D": 4500,
            "E": 5400,
            "F": 10000,
            "G": 10000,
            "H": 10000,
            "J": 7200,
            "K": 7200,
            "L": 7200,
            "M": 5400,
            "N": 5400,
            "P": 7200,
            "R": 10000,
            "S": 7200,
            "T": 7200,
            "U": 5400,
            "V": 5400,
            "W": 3600,
            "X": 4200,
            "Y": 7200,
            "Z": 7200,
        }.get(rpm_num)
        _add_feature_if_possible(disk, "spin-rate-rpm", rpm)

        interface = {
            "A": "ide-ports-n",
            "B": "ide-ports-n",
            "C": "firewire-ports-n",
            "D": "sata-ports-n",
            "E": "ide-ports-n",
            "R": "sata-ports-n",
            "S": "sata-ports-n",
        }.get(interface_num)
        _add_interface_if_possible(disk, interface)

    elif re.match(new_format, model):
        result = re.search(f"^{new_format}", model)
        ff_num = result.group(1)
        rpm_num = result.group(2)
        interface_num = result.group(3)

        ff = {
            "A": ("3.5", None),
            "B": ("2.5", None),
            "C": ("1.0", None),
            "E": ("3.5", None),
            "F": ("3.5", None),
            # "G": "",
            # "H": "",
            "J": ("2.5", 9.5),
            "K": ("3.5", None),
            "L": ("2.5", 7),
            "M": ("2.5", 5),
            "N": ("2.5", 15),
            "P": ("3.5", None),
            "S": ("2.5", 7),
            "T": ("2.5", 12.5),
            "X": ("2.5", 9.5),
        }.get(ff_num)
        if ff:
            _add_feature_if_possible(disk, "hdd-form-factor", ff[0])
            _add_feature_if_possible(disk, "height-mm", ff[1])

        rpm = {
            "A": 5400,
            "B": 7200,
            "C": 5400,
            "D": 5400,
            "E": 7200,
            "F": 10000,
            "G": 10000,
            "H": 10000,
            "J": 7200,
            "K": 7200,
            "L": 7200,
            "P": 5400,
            "R": 5400,
            "S": 7200,
            "T": 10000,
            "V": 5400,
            "W": 7200,
            "Y": 7200,
            "Z": 5400,
        }.get(rpm_num)
        _add_feature_if_possible(disk, "spin-rate-rpm", rpm)

        interface = {
            "A": "ide-ports-n",
            "B": "ide-ports-n",
            "D": "sata-ports-n",
            "E": "ide-ports-n",
            "F": "sas-sata-ports-n",
            "G": "sas-sata-ports-n",
            "K": "sata-ports-n",
            "S": "sata-ports-n",
            "T": "sata-ports-n",
            "W": "usb-ports-n",
            "X": "sata-ports-n",
            "Z": "sata-ports-n",
        }.get(interface_num)
        _add_interface_if_possible(disk, interface)


def maxtor_model_decode(disk, model: str):
    the_format = r"([0-9][A-Z])[0-9]{2,3}([A-Z])[0-9]"

    if re.match(the_format, model):
        result = re.search(f"^{the_format}", model)
        series_num = result.group(1)
        interface_num = result.group(2)

        ff = {
            "4D": "3.5",
            "4G": "3.5",
            "4K": "3.5",
            "6L": "3.5",
            "6E": "3.5",
            "6Y": "3.5",
        }.get(series_num)
        _add_feature_if_possible(disk, "hdd-form-factor", ff)

        interface = {
            "D": "ide-ports-n",
            "F": "sata-ports-n",
            "H": "ide-ports-n",
            "J": "ide-ports-n",
            "K": "ide-ports-n",
            "L": "ide-ports-n",
            "M": "sata-ports-n",
            "P": "ide-ports-n",
            "R": "ide-ports-n",
            "S": "sata-ports-n",
            "U": "ide-ports-n",
        }.get(interface_num)
        _add_interface_if_possible(disk, interface)
    else:
        seagate_model_decode(disk, model, True)


def _add_feature_if_possible(disk, feature, value):
    if value is not None:
        if feature not in disk:
            if value:
                disk[feature] = value


def _add_interface_if_possible(disk, interface):
    interface_values = [
        "ide-ports-n",
        "sata-ports-n",
        "firewire-ports-n",
        "sas-sata-ports-n",
        "usb-ports-n",
        "scsi-sca2-ports-n",
        "scsi-db68-ports-n",
    ]

    if interface:
        if interface not in disk:
            for maybe_interface in interface_values:
                if maybe_interface in disk:
                    # TODO: print a warning
                    break
            else:
                disk[interface] = 1


def parse_single_disk(smartctl: dict, interactive: bool = False) -> dict:
    """
    Parse a single disk from smartctl -ja output to tarallo upload format.

    See parse_smartctl to parse multiple disks.
    """
    disk = {
        "type": "hdd",
    }

    # json_format_version is [1,0], anything else and this parser will catch fire

    port = None

    if smartctl.get("vendor") and smartctl.get("product"):
        # For SCSI disks only, apparently
        disk["brand"] = smartctl.get("vendor")
        disk["model"] = smartctl.get("product")
    else:
        # "Device Model:" is model_name in the JSON
        if smartctl.get("model_name"):
            brand, model = _split_brand_and_other(smartctl.get("model_name"))
            disk["model"] = model
            if "brand" not in disk and brand:
                disk["brand"] = brand

        if smartctl.get("model_family"):
            brand, family = _split_brand_and_other(smartctl.get("model_family"))
            disk["family"] = family
            if "brand" not in disk and brand:
                disk["brand"] = brand

    if disk.get("brand", "") == "WDC":
        disk["brand"] = "Western Digital"

    if smartctl.get("serial_number"):
        disk["sn"] = smartctl.get("serial_number")

    if smartctl.get("wwn"):
        disk["wwn"] = str(smartctl["wwn"].get("naa", "")) + " " + str(smartctl["wwn"].get("oui", "")) + " " + str(smartctl["wwn"].get("id", ""))

    if smartctl.get("form_factor", {}).get("name"):
        ff = smartctl["form_factor"]["name"]
        # https://github.com/smartmontools/smartmontools/blob/master/smartmontools/ataprint.cpp#L405
        if ff == "3.5 inches":
            disk["hdd-form-factor"] = "3.5"
        elif ff == "2.5 inches":
            disk["hdd-form-factor"] = "2.5"
        elif ff == "1.8 inches":
            disk["hdd-form-factor"] = "1.8"
        # TODO: add these to tarallo
        elif ff == "M.2":
            disk["hdd-form-factor"] = "m2"
            port = PORT.M2
        elif ff == "mSATA":
            disk["hdd-form-factor"] = "msata"
            port = PORT.MSATA

    if smartctl.get("user_capacity", {}).get("bytes"):
        # https://stackoverflow.com/a/3411435
        round_digits = int(floor(log10(abs(float(smartctl["user_capacity"]["bytes"]))))) - 2
        bytes_rounded = int(round(float(smartctl["user_capacity"]["bytes"]), -round_digits))
        disk["capacity-decibyte"] = bytes_rounded

    # This may be 0, which is a valid value and casts to False, check for None explicitly!
    if smartctl.get("rotation_rate") is not None:
        if smartctl.get("rotation_rate") > 0:
            disk["spin-rate-rpm"] = smartctl.get("rotation_rate")
            disk["type"] = "hdd"
        else:
            disk["type"] = "ssd"

    if disk.get("brand", "").title() == "Western Digital":
        # These are useless and usually not even printed on labels and in bar codes...
        if "model" in disk:
            disk["model"] = _remove_prefix("WDC ", disk["model"])
        if "sn" in disk:
            disk["sn"] = _remove_prefix("WD-", disk["sn"])

    if "SSD " in disk.get("model", "") or " SSD" in disk.get("model", ""):
        disk["model"] = disk["model"].replace("SSD ", "").replace(" SSD", "")
        _mega_clean_disk_model(disk)
        if disk["type"] == "hdd":
            disk["type"] = "ssd"

    if "(SATA)" in disk.get("family", ""):
        disk["family"] = disk.get("family").replace("(SATA)", "").strip()

    if "(ATA/133 and SATA/150)" in disk.get("family", ""):
        disk["family"] = disk.get("family").replace("(ATA/133 and SATA/150)", "").strip()

    if "SSD" in disk.get("family", ""):
        if disk["type"] == "hdd":
            disk["type"] = "ssd"
        lowered = disk["family"].replace(" ", "").lower()
        if lowered in ("basedssds", "basedssd"):
            del disk["family"]

    if disk.get("family", "").startswith("/"):
        disk["family"] = disk.get("family", "")[1:]

    if "Serial ATA" in disk.get("family", ""):
        if not port:
            port = PORT.SATA
        disk["family"] = disk.get("family").replace("Serial ATA", "").strip()

    if disk.get("model", "").startswith("HGST "):
        disk["model"] = disk.get("model")[5:]
        disk["brand-manufacturer"] = "HGST"

    # Unreliable port detection as a fallback
    if port is None:
        if "SATA" in disk.get("family", "") or "SATA" in disk.get("model", ""):
            port = PORT.SATA
        if "Serial ATA" in disk.get("family", ""):
            # disk["family"] = disk["family"].replace("Serial ATA", "").strip()
            port = PORT.SATA
        if "sata_version" in smartctl:
            port = PORT.SATA
        elif "pata_version" in smartctl:
            if disk.get("hdd-form-factor", "").startswith("2.5") or disk.get("hdd-form-factor", "").startswith("1.8"):
                port = PORT.MINIIDE
            else:
                port = PORT.IDE
        if "nvme_version" in smartctl:
            port = PORT.M2
            if disk.get("type", "") == "hdd":
                disk["type"] = "ssd"
            if "hdd-form-factor" not in disk:
                disk["hdd-form-factor"] = "m2"
        if "device" in smartctl and smartctl["device"].get("type", "") == "scsi" and smartctl["device"].get("protocol", "") == "SCSI":
            disk["notes"] = "This is a SCSI disk, however it is not possible to detect the exact connector type. Please set the correct one manually."

    if port is not None:
        disk[port.value] = 1

    # FF detector
    if "hdd-form-factor" not in disk:
        if "desktop" in disk.get("family", "").lower() and port == PORT.SATA:
            disk["hdd-form-factor"] = "3.5"

    if disk.get("model") is not None:
        brand = disk.get("brand")
        if brand == "Western Digital":
            wd_model_decode(disk, disk.get("model"))
        elif brand == "Seagate":
            seagate_model_decode(disk, disk.get("model"))
        elif brand == "Maxtor":
            maxtor_model_decode(disk, disk.get("model"))
        elif brand == "Samsung":
            samsung_model_decode(disk, disk.get("model"))
        elif brand == "Toshiba":
            toshiba_model_decode(disk, disk.get("model"))
        elif brand == "Fujitsu":
            fujitsu_model_decode(disk, disk.get("model"))
        elif brand == "Hitachi":
            hitachi_model_decode(disk, disk.get("model"))
        elif brand == "Quantum":
            quantum_model_decode(disk, disk.get("model"))

    smart, failing_now = extract_smart_data(smartctl)

    status = smart_health_status(smart, failing_now)
    if status:
        if len(smart) < 2 and status == "ok":
            # Nah bro, I'll pass... "ok" with (nearly) no smart data is meaningless
            pass
        else:
            disk["smart-data"] = status
    else:
        if interactive:
            print("Failed to determine HDD health status from SMART data!")
        pass

    return disk


def extract_smart_data(parsed: dict) -> [Dict, bool]:
    """
    Extract SMART attributes and raw values from smartctl -ja output.
    Also returns failing_now value to indicate if any attributes (except temperature)
    is failing now.
    """
    failing_now = False

    smart = {}
    if parsed.get("ata_smart_attributes", {}).get("table"):
        for line in parsed["ata_smart_attributes"]["table"]:
            # Name
            name = line["name"]

            # Value
            if name.lower() == "unknown_attribute":
                name = f"{name}_{str(line['id'])}"
            value = line["raw"]["value"]

            # Normalize power on time to hours
            # (see https://github.com/mirror/smartmontools/blob/44cdd4ce63ca4e07db87ec062a159181be967a72/ataprint.cpp#L1140-L1168)
            if line["id"] == 9 and name.lower().startswith("power_on_"):
                if "power_on_time" in parsed:
                    if parsed["power_on_time"].get("hours", None) is not None:
                        value = parsed["power_on_time"]["hours"]
                    elif parsed["power_on_time"].get("minutes", None) is not None:
                        value = parsed["power_on_time"]["minutes"] * 60
            # Set result
            smart[name] = value

            # Find out if anything is failing
            if "when_failed" in line:
                if line["when_failed"] == "now":
                    if line["name"].lower() != "temperature_celsius":
                        failing_now = True
    return smart, failing_now


def _mega_clean_disk_model(disk: dict):
    disk["model"] = disk["model"].replace("  ", " ").strip()
    if disk["model"] == "":
        del disk["model"]


def smart_health_status(smart: dict, failing_now: bool) -> str:
    """
    Get disk status from smartctl output.
    This algorithm has been mined: it's based on a decision tree with "accuracy" criterion since seems to produce
    slightly better results than the others. And the tree is somewhat shallow, which makes the algorithm more
    human-readable. There's no much theory other than that, so there's no real theory here.
    The data is about 200 smartctl outputs for every kind of hard disk, manually labeled with pestello (and mortaio)
    according to how I would classify them or how they are acting: if an HDD is making horrible noises and cannot
    perform a single read without throwing I/O errors, it's failed, no matter what the smart data says.
    Initially I tried to mix SSDs in, but their attributes are way different and they are also way easier to
    classify, so this algorithm works on mechanical HDDs only.
    This is the raw tree as output by RapidMiner:
    Current_Pending_Sector > 0.500
    |   Load_Cycle_Count = ?: FAIL {FAIL=9, SUS=0, OK=1, OLD=0}
    |   Load_Cycle_Count > 522030: SUS {FAIL=0, SUS=3, OK=0, OLD=0}
    |   Load_Cycle_Count ≤ 522030: FAIL {FAIL=24, SUS=0, OK=1, OLD=0}
    Current_Pending_Sector ≤ 0.500
    |   Reallocated_Sector_Ct = ?: OK {FAIL=1, SUS=0, OK=4, OLD=0}
    |   Reallocated_Sector_Ct > 0.500
    |   |   Reallocated_Sector_Ct > 3: FAIL {FAIL=8, SUS=1, OK=0, OLD=0}
    |   |   Reallocated_Sector_Ct ≤ 3: SUS {FAIL=0, SUS=4, OK=0, OLD=0}
    |   Reallocated_Sector_Ct ≤ 0.500
    |   |   Power_On_Hours = ?
    |   |   |   Run_Out_Cancel = ?: OK {FAIL=0, SUS=1, OK=3, OLD=1}
    |   |   |   Run_Out_Cancel > 27: SUS {FAIL=0, SUS=2, OK=0, OLD=0}
    |   |   |   Run_Out_Cancel ≤ 27: OK {FAIL=1, SUS=0, OK=6, OLD=1}
    |   |   Power_On_Hours > 37177.500
    |   |   |   Spin_Up_Time > 1024.500
    |   |   |   |   Power_Cycle_Count > 937.500: SUS {FAIL=0, SUS=1, OK=0, OLD=1}
    |   |   |   |   Power_Cycle_Count ≤ 937.500: OK {FAIL=0, SUS=0, OK=3, OLD=0}
    |   |   |   Spin_Up_Time ≤ 1024.500: OLD {FAIL=0, SUS=0, OK=2, OLD=12}
    |   |   Power_On_Hours ≤ 37177.500
    |   |   |   Start_Stop_Count = ?: OK {FAIL=0, SUS=0, OK=3, OLD=0}
    |   |   |   Start_Stop_Count > 13877: OLD {FAIL=1, SUS=0, OK=0, OLD=2}
    |   |   |   Start_Stop_Count ≤ 13877: OK {FAIL=2, SUS=9, OK=89, OLD=4}
    but some manual adjustments were made, just to be safe.
    Most HDDs are working so the data is somewhat biased, but there are some very obvious red flags like smartctl
    reporting failing attributes (except temperature, which doesn't matter and nobody cares) or having both
    reallocated AND pending sectors, where nobody would keep using that HDD, no matter what the tree decides.

    :param failing_now: If any attribute is marked as failing
    :param smart: Smartctl data
    :return: HDD status (label)
    """
    # Oddly the decision tree didn't pick up this one, but it's a pretty obvious sign the disk is failed
    if failing_now:
        return "fail"

    if int(smart.get("Current_Pending_Sector", 0)) > 0:
        # This part added manually just to be safe
        if int(smart.get("Reallocated_Sector_Ct", 0)) > 3:
            return "fail"

        # I wonder if this part is overfitted... who cares, anyway.
        cycles = smart.get("Load_Cycle_Count")
        if cycles:
            if int(cycles) > 522030:
                return "sus"
            else:
                return "fail"
        else:
            return "fail"
    else:
        reallocated = int(smart.get("Reallocated_Sector_Ct", 0))
        if reallocated > 0:
            if reallocated > 3:
                return "fail"
            else:
                return "sus"
        else:
            hours = smart.get("Power_On_Hours")
            if hours:
                # 4.2 years as a server (24/7), 15.2 years in an office pc (8 hours a day, 304 days a year)
                if int(hours) > 37177:
                    if int(smart.get("Spin_Up_Time", 0)) > 1024:
                        # Checking this attribute tells us if it's more likely to be a server HDD or an office HDD
                        if int(smart.get("Power_Cycle_Count", 0)) > 937:
                            # The tree says 1 old and 1 sus here, but there's too little data to throw around "sus"
                            # like this... it needs more investigation, though: if the disk is slow at starting up
                            # it may tell something about its components starting to fail.
                            return "old"
                        else:
                            return "ok"
                    else:
                        return "old"
                else:
                    # This whole area is not very good, but there are too many "ok" disks and too few not-ok ones
                    # to mine something better
                    if int(smart.get("Start_Stop_Count", 0)) > 13877:
                        return "old"
                    else:
                        return "ok"
            else:
                if int(smart.get("Run_Out_Cancel", 0)) > 27:
                    # Fun fact: I never looked at this attribute while classifying HDDs,
                    # but it is indeed a good indication that something is suspicious.
                    return "sus"
                else:
                    return "ok"


def _split_brand_and_other(line):
    lowered = line.lower()

    possibilities = [
        "WDC ",
        "Western Digital",
        "Seagate",
        "Maxtor",
        "Hitachi",
        "Toshiba",
        "Samsung",
        "Fujitsu",
        "Apple",
        "Crucial/Micron",
        "Crucial",
        "LiteOn",
        "Kingston",
        "Adata",
        "Quantum",
    ]

    brand = None
    other = line
    for possible in possibilities:
        if lowered.startswith(possible.lower()):
            brand = possible.strip()
            other = line[len(possible) :].lstrip("_").strip()
            break

    return brand, other


def _remove_prefix(prefix, text):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


if __name__ == "__main__":
    try:
        with open(sys.argv[1], "r") as f:
            input_file = f.read()
        print(json.dumps(parse_smartctl(input_file), indent=2))
    except BaseException as e:
        print(str(e))
        exit(1)



#!/usr/bin/env python3
import copy
import os
import subprocess
import sys
from enum import Enum
from typing import List, Optional, Set

from .constants import basedir
from .parsers.read_decode_dimms import parse_decode_dimms
from .parsers.read_dmidecode import parse_case, parse_motherboard, parse_psu
from .parsers.read_lscpu import parse_lscpu
from .parsers.read_lspci_and_glxinfo import parse_lspci_and_glxinfo
from .parsers.read_smartctl import parse_smartctl


class InputFileNotFoundError(FileNotFoundError):
    def __init__(self, path):
        super().__init__(f"Cannot open file {path}")
        self.path = path

    def get_path(self):
        return self.path


class GenerateFilesError(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class SudoError(GenerateFilesError):
    def __init__(self, msg):
        super().__init__(msg)


MEANINGLESS_VALUES = (
    "",
    "null",
    "custom",
    "unknown",
    "undefined",
    "no enclosure",
    "not available",
    "chassis manufacture",
    "chassis serial number",
    "to be filled by o.e.m",
    "to be filled by o.e.m.",
)


class GpuLocation(Enum):
    NONE = 0
    DISCRETE = 1
    CPU = 2
    MOTHERBOARD = 3


class ParserComponents(Enum):
    CASE = "Case"
    MOTHERBOARD = "Motherboard"
    CPU = "CPU"
    RAM = "RAM"
    GPU = "GPU"
    HDD = "HDD"
    SSD = "SSD"
    PSU = "Power supply"
    ODD = "ODD"
    MONITOR = "Monitor"
    INPUT = "Input devices"

    @classmethod
    def all(cls):
        return list(cls)

    @classmethod
    def all_names(cls):
        res = []
        for thing in cls:
            res.append(thing.value)
        return res

    @classmethod
    def all_keys(cls):
        res = []
        for thing in cls:
            res.append(thing.name)
        return res

    @classmethod
    def not_implemented_yet(cls):
        return {
            ParserComponents.ODD,
            ParserComponents.MONITOR,
            ParserComponents.INPUT,
        }


def check_dependencies_for_generate_files():
    retval = os.system("dpkg -s pciutils i2c-tools mesa-utils smartmontools dmidecode > /dev/null")
    return retval == 0


def generate_files(path: str, use_sudo: bool = True, sudo_passwd: str = None):
    if os.path.exists(os.path.join(basedir, "scripts/generate_files.pxec")):
        script = "scripts/generate_files.pkexec"
    else:
        script = "scripts/generate_files.sh"
    script = os.path.join(basedir, script)
    os.makedirs(path, exist_ok=True)
    command = [script, path]
    if use_sudo:
        command = ["sudo", "-S"] + command

    p = subprocess.Popen(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE, stdin=subprocess.PIPE)

    if sudo_passwd is not None:
        try:
            # out, err =
            p.communicate(input=(sudo_passwd + "\n").encode(), timeout=30)
        except subprocess.TimeoutExpired:
            p.kill()
            raise GenerateFilesError(" ".join(command) + " timed out after 30 seconds")
    else:
        try:
            p.communicate(timeout=30)
        except subprocess.TimeoutExpired:
            p.kill()
            raise GenerateFilesError(" ".join(command) + " timed out after 30 seconds")

    if p.returncode is None:
        raise GenerateFilesError(" ".join(command) + " did not run")
    elif p.returncode != 0:
        if use_sudo and p.returncode == 1:
            raise SudoError(" ".join(command) + f" failed, return code: {p.returncode}")
        else:
            raise GenerateFilesError(" ".join(command) + f" failed, return code: {p.returncode}")

    return path


def required_files():
    return (
        "baseboard.txt",
        "chassis.txt",
        "connector.txt",
        "dimms.txt",
        "glxinfo.txt",
        "lscpu.txt",
        "lspci.txt",
        "net.txt",
        "smartctl.txt",
    )


def _merge_gpu(current_results: List[dict], target_type: str, gpus: list) -> None:
    if len(gpus) <= 0:
        return
    features = _extract_gpu_for_integrated(gpus[0])

    for target in _find_all_components(target_type, current_results):
        target.update(features)


def _extract_gpu_for_integrated(gpu: dict) -> dict:
    if "brand-manufacturer" in gpu and len("brand-manufacturer") > 0:
        brand = gpu["brand-manufacturer"]
    elif "brand" in gpu and len("brand") > 0:
        brand = gpu["brand"]
    else:
        brand = None

    internal_name_present = "internal-name" in gpu
    model_present = "model" in gpu
    if model_present and internal_name_present:
        model = f"{gpu['model']} ({gpu['internal-name']})"
    elif model_present:
        model = gpu["model"]
    elif internal_name_present:
        model = gpu["internal-name"]
    else:
        model = None

    result = {}
    if brand is not None:
        result["integrated-graphics-brand"] = brand
    if model is not None:
        result["integrated-graphics-model"] = model

    return result


def _find_component(component_type: str, result: List[dict]) -> Optional[dict]:
    for component in result:
        if "type" in component and component["type"] == component_type:
            return component
    return None


def _find_all_components(component_type: str, result: List[dict]) -> List[dict]:
    return_this = []
    for component in result:
        if "type" in component and component["type"] == component_type:
            return_this.append(component)
    return return_this


def call_parsers(
    generated_files_path: str,
    components: Set[ParserComponents],
    gpu_location: GpuLocation,
    interactive: bool = False,
) -> list:
    generated_files_path = generated_files_path.rstrip("/")

    def read_file(name: str) -> str:
        path = os.path.join(generated_files_path, name)
        try:
            with open(path, "r") as f:
                output = f.read()
            return output
        except FileNotFoundError:
            raise InputFileNotFoundError(path)

    result = []

    # TODO: if linux, else windows
    if sys.platform == "win32":
        pass
    else:
        if not components.isdisjoint({ParserComponents.CASE, ParserComponents.MOTHERBOARD}):
            result += parse_motherboard(
                read_file("baseboard.txt"),
                read_file("connector.txt"),
                read_file("net.txt"),
                interactive,
            )
            if gpu_location == GpuLocation.MOTHERBOARD:
                _merge_gpu(
                    result,
                    "motherboard",
                    parse_lspci_and_glxinfo(False, read_file("lspci.txt"), ""),
                )
        if ParserComponents.CASE in components:
            result += parse_case(read_file("chassis.txt"), _find_component("motherboard", result))
        if ParserComponents.CPU in components:
            result += parse_lscpu(read_file("lscpu.txt"))
            if gpu_location == GpuLocation.CPU:
                _merge_gpu(
                    result,
                    "cpu",
                    parse_lspci_and_glxinfo(False, read_file("lspci.txt"), ""),
                )
        if ParserComponents.GPU in components and gpu_location == GpuLocation.DISCRETE:
            result += parse_lspci_and_glxinfo(True, read_file("lspci.txt"), read_file("glxinfo.txt"), interactive)
        if ParserComponents.RAM in components:
            result += parse_decode_dimms(read_file("dimms.txt"), interactive)
        if ParserComponents.HDD in components or ParserComponents.SSD in components:
            result += parse_smartctl(read_file("smartctl.txt"), interactive)
        if ParserComponents.PSU in components:
            result += parse_psu(_find_component("case", result))

    result = _do_cleanup(result, interactive)
    return result


def split_products(parsed: List[dict]) -> List[dict]:
    item_keys = item_only_features()
    bmv = [
        "brand",
        "model",
        "variant",
    ]
    both = [
        "type",
    ]

    final_result = []
    products = []

    for item in parsed:
        if can_be_product(item):
            if item.get("variant", "") == "":
                item["variant"] = "default"
            new_product = {k: item.get(k) for k in bmv if k in item.keys()}
            found = False
            for old_product in products:
                if new_product.items() <= old_product.items():
                    found = True
                    break
            if not found:
                new_product.update(
                    {
                        "type": "P",
                        "features": {k: v for k, v in item.items() if k not in bmv + item_keys},
                    }
                )
                products.append(new_product)
        new_item = {
            "type": "I",
            "features": {k: v for k, v in item.items() if k in bmv + both + item_keys},
            "contents": [],
        }
        final_result.append(new_item)

    final_result += products
    return final_result


def item_only_features():
    item_keys = [
        "arrival-batch",
        "cib",
        "cib-old",
        "cib-qr",
        "data-erased",
        "mac",
        "notes",
        "os-license-code",
        "os-license-version",
        "other-code",
        "owner",
        "smart-data",
        "sn",
        "software",
        "surface-scan",
        "working",
        "wwn",
    ]
    return item_keys


def add_owner(parsed_items_only: List[dict], owner: str) -> List[dict]:
    for item in parsed_items_only:
        item["owner"] = owner
    return parsed_items_only


def add_chassis_code(parsed: List[dict], code: str) -> bool:
    for item in parsed:
        if item.get("type") == "I":
            if item.get("features", {}).get("type") == "case":
                item["code"] = code
                return True
    return False


def can_be_product(component: dict):
    # check if brand and model exist
    if "brand" not in component or "model" not in component:
        return False

    return True


def _do_cleanup(result: List[dict], verbose: bool = False) -> List[dict]:
    by_type = {}

    for item in result:
        removed = set()
        for k, v in item.items():
            # Check for k in these?
            # "brand",
            # "brand-manufacturer",
            # "model",
            # "integrated-graphics-brand",
            # "integrated-graphics-model",
            if isinstance(v, str) and v.lower() in MEANINGLESS_VALUES:
                removed.add(k)
            elif v is None:
                removed.add(k)
        for removed_thing in removed:
            del item[removed_thing]
        the_type = item.get("type")
        if the_type not in by_type:
            by_type[the_type] = []
        by_type[the_type].append(item)

        if verbose and len(removed) > 0:
            print(f"WARNING: Removed from {item.get('type', 'item with no type')}: {', '.join(removed)}.")

    for case in by_type.get("case", []):
        for mobo in by_type.get("motherboard", []):
            try:
                if (case["model"], case["brand"], case["variant"]) == (
                    mobo["brand"],
                    mobo["model"],
                    mobo["variant"],
                ):
                    case.pop("model")
            except KeyError:
                pass

    # avoid bad associations between items and products
    if len(result) > 1:
        for component1 in result:
            i = 1
            for component2 in result[i:]:
                if component1["type"] != component2["type"]:
                    if can_be_product(component1) and can_be_product(component2):
                        if (component1["brand"], component2["model"]) == (
                            component2["brand"],
                            component2["model"],
                        ):
                            variant1 = component1.get("variant", "")
                            variant2 = component2.get("variant", "")
                            if variant1 == variant2:
                                component1["variant"] = variant1.rstrip().join(f"_{component1['type']}").lstrip("_")
                                component2["variant"] = variant2.rstrip().join(f"_{component2['type']}").lstrip("_")

    return result


def _should_be_in_motherboard(the_type: str, features: dict) -> bool:
    if the_type in ("cpu", "ram"):
        return True
    if the_type.endswith("-card"):
        return True
    if the_type == "ssd":
        if features.get("hdd-form-factor", None) in ("m2", "m2.2"):
            return True
    return False


def _should_be_in_case(the_type: str, features: dict) -> bool:
    if the_type in ("motherboard", "hdd", "ssd", "odd", "fdd", "psu"):
        return True
    # Fallback for when there's no motherboard
    return _should_be_in_motherboard(the_type, features)


def unmake_tree(items_and_products: List[dict]) -> List[dict]:
    result = []

    for thing in items_and_products:
        result.append(thing)
        if thing.get("type") == "I":
            if "contents" in thing:
                result += unmake_tree(thing["contents"])
                thing["contents"] = []
    return result


def make_tree(items_and_products: List[dict]) -> List[dict]:
    items_and_products = copy.deepcopy(items_and_products)
    by_type = {}
    products = []

    for thing in items_and_products:
        if thing.get("type") == "I":
            if "features" in thing:
                the_type = thing["features"].get("type")
                if the_type not in by_type:
                    by_type[the_type] = []
                by_type[the_type].append(thing)
                continue

        products.append(thing)

    if "motherboard" in by_type:
        containers = by_type["motherboard"]
        del by_type["motherboard"]
        for the_type in by_type:
            save = []
            for thing in by_type[the_type]:
                if _should_be_in_motherboard(the_type, thing.get("features", {})):
                    if "contents" not in containers[0]:
                        containers[0]["contents"] = []
                    containers[0]["contents"].append(thing)
                else:
                    save.append(thing)
            by_type[the_type] = save
        save2 = {}
        for the_type in by_type:
            if len(by_type[the_type]) > 0:
                save2[the_type] = by_type[the_type]
        by_type = save2
        by_type["motherboard"] = containers

    if "case" in by_type:
        containers = by_type["case"]
        del by_type["case"]
        for the_type in by_type:
            save = []
            for thing in by_type[the_type]:
                if _should_be_in_case(the_type, thing.get("features", {})):
                    if "contents" not in containers[0]:
                        containers[0]["contents"] = []
                    containers[0]["contents"].append(thing)
                else:
                    save.append(thing)
                by_type[the_type] = save
        save2 = {}
        for the_type in by_type:
            if len(by_type[the_type]) > 0:
                save2[the_type] = by_type[the_type]
        by_type = save2
        by_type["case"] = containers

    top_items = []
    for the_type in by_type:
        top_items += by_type[the_type]

    return top_items + products


def check_required_files(path, is_gui: bool = False):
    if os.path.isdir(path):
        files_in_dir = os.listdir(path)
        if not files_in_dir:
            return ""
        for file in required_files():
            for file_in_dir in files_in_dir:
                if file_in_dir == file:
                    break
            else:
                if is_gui:
                    error = f"Missing file {file}\n"
                    return error
                else:
                    print(f"[bold red]Missing file {file}\n" f"Please re-run this script without the -f or --files option.[/]")
                    exit(1)
    return ""


def env_to_bool(value: Optional[str]) -> bool:
    try:
        if value.lower() in ("1", "true", "t", "", "yes", "y"):
            return True
    except AttributeError:
        pass

    return False



import os
import signal
import sys

from PyQt6 import QtWidgets

from . import peracruda
from .commons import env_to_bool
from .config import CONFIG
from .constants import VERSION
from .gui import GUI, gui_excepthook
from .peralog import logdir, logger


def common_args_parsing():
    """Parse arguments common to both GUI and CLI version
    --version prints the current version and quits.
    --logs prints the path where logs are stored and quits.
    """
    if any([s in sys.argv for s in ["--version", "-v"]]):
        print(f"P.E.R.A.C.O.T.T.A. Version {VERSION}")
        exit(0)

    if "--logs" in sys.argv:
        print(f"P.E.R.A.C.O.T.T.A.'s logs are located in {logdir.as_posix()}")
        exit(0)


def main_gui():
    common_args_parsing()

    app = QtWidgets.QApplication(sys.argv)
    sys.excepthook = gui_excepthook
    # makes CTRL+ C work
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # noinspection PyUnusedLocal
    window = GUI(app)
    app.exec()
    logger.info("Started PERACOTTA")


def main_cli():
    print("Sorry, peracruda isn't implemented in v2 yet! Use the old one at https://github.com/WEEE-Open/peracotta")
    common_args_parsing()
    # peracruda.main_()



import json
import os
import shutil
import sys
import time
import urllib.parse
import urllib.request
from collections import defaultdict

from PyQt6 import QtCore, QtGui, QtWidgets, uic
import requests

from peracotta import commons
from peracotta.commons import ParserComponents, make_tree
from peracotta.config import conf_dir, CONFIG
from peracotta.constants import ICON, PATH, URL, VERSION
from peracotta.tarallo import TaralloUploadDialog, Uploader, tarallo_success_dialog
from peracotta.peralog import logger

from .widgets import JsonWidget, ErrorDialog
from .exceptions import MissingFeaturesError

from .PeraThread import PeracottaThread
from .Toolbox import ToolBoxWidget


DEFAULT_PROGRESS_BAR_STYLE = (
    "QStatusBar::item {"
    "min-height: 12px;"
    "max-height: 12px;"
    "}"
    "QProgressBar {"
    "min-height: 14px;"
    "max-height: 14px;"
    "}"
    "QProgressBar::chunk {"
    "background-color: #00963A;"
    "width: 50px;"
    "}"
)


class GUI(QtWidgets.QMainWindow):
    def __init__(
        self,
        app: QtWidgets.QApplication,
    ) -> None:
        super(GUI, self).__init__()
        uic.loadUi(PATH["UI"], self)
        self.app = app
        self.uploader = None
        self.taralloDialog = None
        self.data = list(dict())
        self.features = dict()
        self.encountered_types_count = defaultdict(lambda: 0)
        self.active_theme = str()

        self.setWindowIcon(QtGui.QIcon(PATH["ICON"]))

        # shortcuts
        self.refreshThemeShortcut = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+R"), self)
        self.refreshThemeShortcut.activated.connect(self.refresh_theme)

        # Output toolbox
        self.outputScrollArea = self.findChild(QtWidgets.QScrollArea, "outputScrollArea")
        self.itemToolBox = None

        # App settings
        self.settings: QtCore.QSettings = QtCore.QSettings("WEEE Open", "PERACOTTA")

        # Gpu location layout
        self.gpuGroupBox = self.findChild(QtWidgets.QGroupBox, "gpuGroupBox")

        # Radio buttons
        self.discreteRadioBtn = self.findChild(QtWidgets.QRadioButton, "discreteRadioBtn")
        self.intCpuRadioBtn = self.findChild(QtWidgets.QRadioButton, "intCpuRadioBtn")
        self.intMoboRadioBtn = self.findChild(QtWidgets.QRadioButton, "intMoboRadioBtn")
        self.bothGpuRadioBtn = self.findChild(QtWidgets.QRadioButton, "bothGpuRadioBtn")

        # Selectors area
        self.selectorsWidget = self.findChild(QtWidgets.QWidget, "selectorsWidget")
        self.selectorsScrollArea = self.findChild(QtWidgets.QScrollArea, "selectorsScrollArea")

        self.addItemComboBox = self.findChild(QtWidgets.QComboBox, "addItemComboBox")
        self.addItemComboBox.addItem("Select Type --")
        self.addItemComboBox.currentTextChanged.connect(self.add_toolbox_item)
        self.addItemComboBox.wheelEvent = lambda stop_wheel_event: None

        # 'select/deselect all' buttons
        self.selectAllBtn = self.findChild(QtWidgets.QPushButton, "selectBtn")
        self.selectAllBtn.clicked.connect(self.select_all_checkboxes)
        self.deselectAllBtn = self.findChild(QtWidgets.QPushButton, "deselectBtn")
        self.deselectAllBtn.clicked.connect(self.deselect_all_checkboxes)

        # Owner line edit
        self.ownerLineEdit = self.findChild(QtWidgets.QLineEdit, "ownerLineEdit")

        # Generate data button
        self.generateBtn = self.findChild(QtWidgets.QPushButton, "generateBtn")
        self.generateBtn.clicked.connect(self.generate)

        # Reset selectors button
        self.resetBtn = self.findChild(QtWidgets.QPushButton, "resetBtn")
        self.resetBtn.clicked.connect(self.reset_setup_group)

        # Save JSON button
        self.saveJsonBtn = self.findChild(QtWidgets.QPushButton, "saveJsonBtn")
        self.saveJsonBtn.clicked.connect(self.save_json)

        # Upload to tarallo button
        self.uploadBtn = self.findChild(QtWidgets.QPushButton, "uploadBtn")
        self.uploadBtn.clicked.connect(self.tarallo_dialog)

        # File actions
        self.actionOpen = self.findChild(QtGui.QAction, "actionOpen")
        self.actionOpen.triggered.connect(self.open_json)
        self.actionOpenLastJson = self.findChild(QtGui.QAction, "actionOpenLastJson")
        self.actionOpenLastJson.triggered.connect(self.open_latest_json)
        self.actionOpenJson = self.findChild(QtGui.QAction, "actionOpenJson")
        self.actionOpenJson.triggered.connect(self.show_json)
        self.actionLoadRawFiles = self.findChild(QtGui.QAction, "actionLoadRawFiles")
        self.actionLoadRawFiles.triggered.connect(self._load_raw_files)
        self.actionExit = self.findChild(QtGui.QAction, "actionExit")
        self.actionExit.triggered.connect(self.close)

        # Options actions
        self.menuTheme = self.findChild(QtWidgets.QMenu, "menuTheme")
        action = list()
        action.append(self.menuTheme.addAction("Default"))
        action[-1].triggered.connect(lambda: self.set_theme("default"))
        for theme_file in os.listdir(PATH["THEMES"]):
            theme = theme_file.rstrip(".css")
            action.append(self.menuTheme.addAction(theme))
            action[-1].triggered.connect((lambda t: lambda: self.set_theme(t))(theme))

        # Help actions
        self.actionAboutUs = self.findChild(QtGui.QAction, "actionAboutUs")
        self.actionAboutUs.triggered.connect(self.open_website)
        self.actionSourceCode = self.findChild(QtGui.QAction, "actionSourceCode")
        self.actionSourceCode.triggered.connect(self.open_source_code)
        self.actionVersion = self.findChild(QtGui.QAction, "actionVersion")
        self.actionVersion.triggered.connect(self.show_version)

        # Status bar widgets
        self.progressBar = QtWidgets.QProgressBar()
        self.statusBar().addPermanentWidget(self.progressBar)
        self.progressBar.hide()
        # self.statusBar().setStyleSheet(DEFAULT_PROGRESS_BAR_STYLE)

        # Setup peracotta QThread
        self.perathread = PeracottaThread(self)
        self.perathread.updateEvent.connect(self.peracotta_results)
        self.perathread.startEvent.connect(self.show_busy_progress_bar)
        self.perathread.errorEvent.connect(self.peracotta_error)

        self.errorDialog = None

        self.show()
        self.setup()

    def setup(self):
        self.set_theme(self.settings.value("last_theme", "default"))
        self.load_features_file(CONFIG["TARALLO_FEATURES_AUTO_DOWNLOAD"])

        # Set item types available in the add item combo box
        for type_key in self.features["values"]["type"]:
            type_value = self.features["values"]["type"][type_key]
            self.addItemComboBox.addItem(type_value)
            if type_key in ICON:
                icon = QtGui.QIcon(ICON[type_key])
                self.addItemComboBox.setItemIcon(self.addItemComboBox.count() - 1, icon)

        # Set up the item toolbox
        self.itemToolBox = ToolBoxWidget(self.data, self.features, self.encountered_types_count)
        self.outputScrollArea.setWidget(self.itemToolBox)

        self.reset_toolbox()

        # Set the selectors widget
        layout = self.selectorsWidget.layout()
        niy = ParserComponents.not_implemented_yet()
        for item in ParserComponents:
            checkbox = QtWidgets.QCheckBox(item.value)
            if item in niy:
                checkbox.setEnabled(False)
            layout.addWidget(checkbox)
        self.reset_setup_group()

    @staticmethod
    def backup_features_json():
        shutil.copy2(conf_dir.joinpath("features.json"), conf_dir.joinpath("features.json.bak"))

    @staticmethod
    def restore_features_json():
        shutil.move(conf_dir.joinpath("features.json.bak"), conf_dir.joinpath("features.json"))

    def load_features_file(self, auto_update: bool):
        self.features = {}
        has_file = False

        try:
            mtime = os.path.getmtime(PATH["FEATURES"])
            self.backup_features_json()
            has_file = True
        except FileNotFoundError:
            mtime = 0

        if auto_update and time.time() - mtime > 60 * 60 * 12:
            try:
                response = requests.get(f"{CONFIG['TARALLO_URL']}/features.json", headers={"User-Agent": "peracotta", "Accept": "application/json"})

                with open(conf_dir.joinpath("features.json"), "wb") as fs:
                    json.dump(response.json(), fs)

                has_file = True
            except requests.exceptions.ConnectionError as e:
                logger.exception("Couldn't connect to TARALLO")
                QtWidgets.QMessageBox.warning(None, "Error", f"Couldn't connect to TARALLO to update features.json")
            except Exception as e:
                logger.exception(e)
            finally:
                if not has_file:
                    try:
                        self.restore_features_json()
                        has_file = True
                    except FileNotFoundError as e:
                        pass

        if has_file:
            self.parse_features_file()
        else:
            raise MissingFeaturesError("features.json file not present")

    def parse_features_file(self):
        with open(PATH["FEATURES"], "r") as file:
            feature_names = {}
            feature_types = {}
            feature_values = {}
            features = json.load(file)
            for group in features["features"]:
                for feature in features["features"][group]:
                    name = feature["name"]
                    feature_names[name] = feature["printableName"]
                    feature_types[name] = feature["type"]
                    if "values" in feature:
                        feature_values[name] = feature["values"]
            self.features = {
                "names": feature_names,
                "types": feature_types,
                "values": feature_values,
            }

    # gui utilities

    def reset_toolbox(self):
        for idx in range(self.itemToolBox.count()):
            self.itemToolBox.removeItem(0)

    def open_url(self, url_type: str):
        url = QtCore.QUrl(url_type)
        if not QtGui.QDesktopServices.openUrl(url):
            QtWidgets.QMessageBox.warning(self, "Cannot Open Url", f"Could not open url {url_type}")

    def set_theme(self, theme: str = "default"):
        logger.debug(f"Setting theme {theme}")
        with open(f"{PATH['THEMES']}{theme}.css", "r") as file:
            self.app.setStyleSheet(file.read())
        self.settings.setValue("last_theme", theme)
        self.active_theme = theme
        logger.debug(f"Done setting theme")

    def refresh_theme(self):
        if self.active_theme == "default":
            return
        with open(f"{PATH['THEMES']}{self.active_theme}.css", "r") as file:
            self.app.setStyleSheet(file.read())

    def show_busy_progress_bar(self):
        self.progressBar.setRange(0, 0)
        self.progressBar.show()

    def select_all_checkboxes(self):
        for checkbox in self.selectorsWidget.findChildren(QtWidgets.QCheckBox):
            if checkbox.isEnabled():
                checkbox.setChecked(True)

    def deselect_all_checkboxes(self):
        for checkbox in self.selectorsWidget.findChildren(QtWidgets.QCheckBox):
            checkbox.setChecked(False)

    def get_file_directory_dialog(self):
        the_dir = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Open JSON",
            f"{os.path.expanduser('~')}",
            f"JSON (*.json);;All Files (*)",
        )
        return the_dir[0]

    def get_directory_dialog(self):
        the_dir = QtWidgets.QFileDialog.getExistingDirectory(self, "Open JSON", f"{os.path.expanduser('~')}")
        return the_dir if the_dir != "" else None

    # tarallo utilities

    def upload_to_tarallo(self, checkbox: bool, bulk_id=None):
        if bulk_id == "":
            bulk_id = None
        self.uploader = Uploader(make_tree(self.data), CONFIG["TARALLO_URL"], CONFIG["TARALLO_TOKEN"], bulk_id, checkbox)
        self.uploader.successEvent.connect(self.tarallo_success)
        self.uploader.failureEvent.connect(self.tarallo_failure)
        self.uploader.start()

    def tarallo_success(self, code: str):
        self.uploader = None
        url = f"{CONFIG['TARALLO_URL']}/bulk/import#{urllib.parse.quote(code)}"
        tarallo_success_dialog(url)

    def tarallo_failure(self, case: str, bulk_id: str):
        self.uploader = None
        if case == "cannot_upload":
            QtWidgets.QMessageBox.warning(
                self,
                "Cannot upload to T.A.R.A.L.L.O.",
                "Cannot upload, try to change the bulk identifier or check the overwrite checkbox.",
            )
            self.tarallo_dialog(bulk_id)
        elif case == "cannot_reach":
            QtWidgets.QMessageBox.warning(
                self,
                "Unable to reach the T.A.R.A.L.L.O.",
                "Please connect this PC to the Internet and try again.",
            )

    # buttons functions

    def reset_setup_group(self):
        # reset gpu location
        for radioBtn in self.gpuGroupBox.findChildren(QtWidgets.QRadioButton):
            radioBtn.setAutoExclusive(False)
            radioBtn.setChecked(False)
            radioBtn.setAutoExclusive(True)

        # reset checkboxes
        defaults = set(commons.ParserComponents.all_names()) - {
            commons.ParserComponents.MONITOR.value,
            commons.ParserComponents.INPUT.value,
        }

        for checkbox in self.selectorsWidget.findChildren(QtWidgets.QCheckBox):
            checkbox: QtWidgets.QCheckBox
            if checkbox.text() in defaults and checkbox.isEnabled():
                checkbox.setChecked(True)
            else:
                checkbox.setChecked(False)
                checkbox.setStyleSheet("text-decoration: line-through;")

        # reset owner
        self.ownerLineEdit.clear()

    def gpu_location_from_buttons(self):
        if self.discreteRadioBtn.isChecked():
            return commons.GpuLocation.DISCRETE
        if self.intCpuRadioBtn.isChecked():
            return commons.GpuLocation.CPU
        if self.intMoboRadioBtn.isChecked():
            return commons.GpuLocation.MOTHERBOARD
        if self.bothGpuRadioBtn.isChecked():
            QtWidgets.QMessageBox.information(
                self,
                "Warning",
                "The integrated GPU cannot be detected in this configuration: disconnect the dedicated one and re-run peracotta if you want to parse it, or edit manually.",
            )
            return commons.GpuLocation.DISCRETE
        else:
            QtWidgets.QMessageBox.warning(self, "Warning", "Please, select one of the GPU locations to proceed.")
            return None

    def get_selected_filters(self):
        filters = set()
        for checkbox in self.selectorsWidget.findChildren(QtWidgets.QCheckBox):
            checkbox: QtWidgets.QCheckBox
            if checkbox.isChecked():
                filters.add(commons.ParserComponents(checkbox.text()))
        return filters

    def generate(self):
        if self.perathread.isRunning():
            return

        if not self.set_thread_buttons_values():
            self.perathread.set_default_values()
            return

        if sys.platform != "win32":
            use_sudo = commons.env_to_bool(os.environ.get("GENERATE_FILES_USE_SUDO", "1"))
            ask_sudo_pass = commons.env_to_bool(os.environ.get("GENERATE_FILES_ASK_SUDO_PASSWORD", "1"))

            if use_sudo and "NOPASSWD: ALL" in os.popen("sudo -l").read():
                ask_sudo_pass = False

            self.perathread.use_sudo = use_sudo

            if use_sudo and ask_sudo_pass:
                got_it = self._ask_sudo_pass()
                if not got_it:
                    return
            else:
                self.perathread.sudo_passwd = None

        self.perathread.generate_files = True
        # TODO: shouldn't the next 2 lines be reversed?
        self.perathread.begin()
        self.reset_toolbox()

    def set_thread_buttons_values(self):
        gpu_location = self.gpu_location_from_buttons()
        if gpu_location is None:
            return False
        self.perathread.gpu_location = gpu_location
        self.perathread.owner = self.ownerLineEdit.text()
        self.perathread.filters = self.get_selected_filters()
        return True

    def _ask_sudo_pass(self):
        sudo_passwd, ok = QtWidgets.QInputDialog.getText(
            self,
            "Insert sudo password",
            "Insert sudo password:",
            QtWidgets.QLineEdit.EchoMode.Password,
        )
        if ok:
            self.perathread.sudo_passwd = sudo_passwd
            return True
        else:
            self.perathread.sudo_passwd = None
            return False

    def save_json(self):
        if self.data is None:
            QtWidgets.QMessageBox.warning(self, "Warning", "There is nothing to be saved")
            return
        the_dir = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save Peracotta JSON",
            f"{os.path.expanduser('~')}",
            "JSON (*.json);;Text file (*.txt);;All Files (*)",
        )
        if the_dir[0] == "":
            return
        with open(the_dir[0], "w") as file:
            file.write(f"{json.dumps(commons.make_tree(self.data), indent=2)}")

    def tarallo_dialog(self, bulk_id=str()):
        if len(self.data) == 0:
            QtWidgets.QMessageBox.warning(self, "Warning", "There is nothing to be uploaded")
            return
        if not bulk_id:
            bulk_id = ""
        taralloDialog = TaralloUploadDialog(self, bulk_id)
        taralloDialog.signal.connect(self.upload_to_tarallo)

    def add_toolbox_item(self):
        if self.addItemComboBox.currentIndex() == 0:
            return
        else:
            item_type = self.addItemComboBox.currentText().lower()
            self.itemToolBox.add_item(item_type=item_type, single_item=True)
            if self.data is None:
                self.data = []
            self.data.append({})
            self.addItemComboBox.setCurrentIndex(0)

    # menu actions

    def open_json(self, path: str = ""):
        if not path:
            path = self.get_file_directory_dialog()
        if path == "":
            self.data = None
            return
        try:
            with open(path, "r") as file:
                self.data = json.load(file)
        except FileNotFoundError as exc:
            QtWidgets.QMessageBox.warning(self, "Error", f"File not found.\n{exc.args[1]}")

        self.data = commons.unmake_tree(self.data)
        self.settings.setValue("latest_json", path)
        self.itemToolBox.load_items(self.data)

    def open_latest_json(self):
        for key in self.settings.childKeys():
            if "latest_json" in key:
                self.open_json(self.settings.value("latest_json"))

    # the checked parameter exists for QAction::triggered
    # noinspection PyUnusedLocal

    def _load_raw_files(self, checked):
        self.load_raw_files()

    def load_raw_files(self, path: str = ""):
        if self.perathread.isRunning():
            return

        if not self.set_thread_buttons_values():
            self.perathread.set_default_values()
            return

        if path == "":
            path = self.get_directory_dialog()
        if path is None:
            self.perathread.set_default_values()
            return
        self.perathread.generate_files = False
        self.perathread.files_path = path

        # TODO: shouldn't the next 2 lines be reversed?
        self.perathread.begin()
        self.reset_toolbox()

    def show_json(self):
        if self.data is None:
            return
        JsonWidget(commons.make_tree(self.data), self.size())

    def open_website(self):
        self.open_url(URL["website"])

    def open_source_code(self):
        self.open_url(URL["source_code"])

    def show_version(self):
        QtWidgets.QMessageBox.about(self, "Version", f"Peracotta v{VERSION}")

    # multithread
    def peracotta_results(self, data: list):
        self.progressBar.setRange(0, 1)  # disable statusBar's progressBar
        self.progressBar.hide()
        if not data:
            return
        self.data = data
        self.itemToolBox.load_items(self.data)

    def peracotta_error(self, error_type: str, error: str):
        self.errorDialog = ErrorDialog(self, error_type, error)
        self.progressBar.setRange(0, 1)  # disable statusBar's progressBar
        self.progressBar.hide()

    # close event
    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        if self.perathread.isRunning():
            self.perathread.terminate()



#!/usr/bin/env python3

import argparse
import json
import os
import random
from datetime import datetime
from os import environ as env
from typing import Optional

from dotenv import load_dotenv
from pytarallo import Tarallo
from pytarallo.Errors import NoInternetConnectionError
from rich import print
from rich.console import Console

from . import commons as peracommon


def main(the_args):
    load_dotenv()
    parsers = []

    if the_args.parsers is None:
        parsers = peracommon.ParserComponents.all()
    else:
        for parser_piece in the_args.parsers.split(","):
            try:
                parsers.append(peracommon.ParserComponents[parser_piece.strip()])
            except KeyError:
                print(f"[red]Unknown component parser: {parser_piece.strip()}[/]")
                print(f"Available parsers: " + ", ".join(peracommon.ParserComponents.all_keys()))
                exit(2)

    if the_args.path is None:
        generated_files_path = f"{os.getcwd()}/tmp"
    else:
        generated_files_path = the_args.path

    # if the_args.files is not None:
    #     peracommon.check_required_files(generated_files_path)

    # If we have to generate files...
    if not the_args.files:
        if os.path.isdir(generated_files_path):
            sel = input(f"Overwrite existing files in {generated_files_path} dir? y/N ").lower()
            if sel == "y":
                print("Overwriting...")
            else:
                sel = input("Output files to current working directory instead? y/N ").lower()
                if sel == "y":
                    generated_files_path = os.getcwd()
                    print("Outputting files to working directory...")
                else:
                    print("[blue]Quitting...[/]")
                    exit(-1)

        # if not generated_files_path:
        #     has_dependencies = peracommon.check_dependencies_for_generate_files()
        #     if not has_dependencies:
        #         if not ask_install_depdendencies():
        #             print("[blue]Quitting...[/]")
        #             exit(-1)

        # now that I have a dest folder, I generate files
        use_sudo = peracommon.env_to_bool(env.get("GENERATE_FILES_USE_SUDO", "1"))
        # ask_sudo_pass = peracommon.env_to_bool(env.get("GENERATE_FILES_ASK_SUDO_PASSWORD", "1"))
        try:
            generated_files_path = peracommon.generate_files(generated_files_path, use_sudo, None)
        except peracommon.GenerateFilesError as e:
            print(f"[red]Error: {str(e)}[/]")
            exit(2)

    gpu_location = get_gpu(the_args)

    # List of items only
    result = []
    try:
        result = peracommon.call_parsers(generated_files_path, set(parsers), gpu_location, True)
    except peracommon.InputFileNotFoundError as e:
        msg = f"[red]Cannot find required file: {e.path}[/]"
        if the_args.files:
            msg += "\nMake sure the file exists or selecting other parsers with -p"
        else:
            msg += "\nMake sure the file exists or try running generate_files.sh manually"
        exit(3)

    owner = get_owner()
    if owner:
        # List of items
        result = peracommon.add_owner(result, owner)

    # List of items and products
    result = peracommon.split_products(result)

    code = get_code()
    if code:
        # List of items
        found = peracommon.add_chassis_code(result, code)
        if not found:
            print("[red]Failed to add code to case! Maybe the case was not parsed?[/]")
            exit(1)

    # List of items as trees and products as products
    result = peracommon.make_tree(result)

    print_output(json.dumps(result, indent=2), generated_files_path)

    prompt_to_open_browser()

    upload_to_tarallo(result)


def ask_install_depdendencies():
    ans = input("You need to install some packages. Do you want to install them? y/N ").lower()
    if ans == "y":
        install_cmd = "apt install -y pciutils i2c-tools mesa-utils smartmontools dmidecode < /dev/null"
        if os.geteuid() != 0:
            os.system(f"sudo {install_cmd}")
        else:
            os.system(f"/bin/bash -c {install_cmd}")
        return True
    else:
        return False


def prompt_to_open_browser():
    import base64

    web_link = "aHR0cHM6Ly90YXJhbGxvLndlZWVvcGVuLml0L2J1bGsvYWRkCg=="
    web_link = base64.b64decode(web_link).decode("ascii").rstrip()
    egg = Console()
    text = [
        "Congratulations!!!",
        "You're",
        "the",
        "1000th",
        "WEEEisitor",
        "of",
        "the",
        "day",
    ]
    this_moment = datetime.now()
    if this_moment.minute == this_moment.second:
        for word in text:
            red = random.randint(0, 255)
            green = random.randint(0, 255)
            blue = random.randint(0, 255)
            egg.print(word, end=" ", style=f"rgb({red},{green},{blue})")
        egg.print(web_link)
    else:
        print(f"[green]Finished successfully![/] Now you can add this output to the T.A.R.A.L.L.O. -> {web_link}")


def upload_to_tarallo(data):
    msg_upload_ok = "[green]All went fine! [/] [blue]\nBye bye! [/]🍐\n"
    msg_upload_failed = "The upload failed. Check above and try to upload on your own"

    ans = input("Do you want to automatically upload the JSON to the T.A.R.A.L.L.O ? (Y/n): ").lower().rstrip()

    if ans.lower() == "n":
        print("\n[blue]Bye bye! [/]🍐\n")

        return

    try:
        t_url = env["TARALLO_URL"]
        t_token = env["TARALLO_TOKEN"]
    except KeyError:
        raise EnvironmentError("Missing definitions of TARALLO* environment variables (see the README)")

    while True:
        try:
            bulk_id = input("Please enter a bulk identifier (optional): ").rstrip()
            if bulk_id == "":
                bulk_id = None
            t = Tarallo.Tarallo(t_url, t_token)
            ver = t.bulk_add(data, bulk_id, False)
            if ver:
                print(msg_upload_ok)
                break
            else:
                overwrite = input("Cannot update, do you want to try overwriting the identifier? (y/N): ").lower().rstrip()
                if overwrite.lower() == "y":
                    ver = t.bulk_add(data, bulk_id, True)
                    if ver:
                        print(msg_upload_ok)
                        break
                    else:
                        print(msg_upload_failed)
                else:
                    bulk_id = input("Do you want to use another identifier? Just press enter for an automatic one. " "You choose (NEW_ID/n): ").rstrip()
                    if bulk_id.lower() != "n":
                        if bulk_id == "":
                            bulk_id = None
                        ver = t.bulk_add(data, bulk_id, True)
                        if ver:
                            print(msg_upload_ok)
                            break
                        else:
                            print(msg_upload_failed)

        except NoInternetConnectionError:
            print("\n[yellow]Unable to reach the T.A.R.A.L.L.O. " "Please connect this PC to the Internet and try again.[/]\n")


def get_gpu(the_args) -> peracommon.GpuLocation:
    # if the_args.files is not None:
    #     the_args.cpu = False
    #     the_args.gpu = False
    #     the_args.motherboard = False
    #
    #     try:
    #         with open(os.path.join(os.getcwd(), the_args.files, "gpu_location.txt")) as f:
    #             location = f.readline().lower().rstrip()
    #             if location == "mobo":
    #                 the_args.motherboard = True
    #             elif location == "gpu":
    #                 the_args.gpu = True
    #             elif location == "cpu":
    #                 the_args.cpu = True
    #     except FileNotFoundError:
    #         pass

    location = None
    if the_args.cpu:
        location = peracommon.GpuLocation.CPU
    elif the_args.gpu:
        location = peracommon.GpuLocation.DISCRETE
    elif the_args.motherboard:
        location = peracommon.GpuLocation.MOTHERBOARD
    elif the_args.gpu_none:
        location = peracommon.GpuLocation.NONE

    while not location:
        print(
            "\nWhere is GPU in your PC? c/g/b/n\n",
            "c for integrated in CPU\n",
            "g for discrete graphics card\n",
            "b for integrated in the motherboard\n",
            "n if there's none\n",
        )
        gpu_flag = input("Insert your choice: ").lower()
        if gpu_flag == "c":
            location = peracommon.GpuLocation.CPU
        elif gpu_flag == "g":
            location = peracommon.GpuLocation.DISCRETE
        elif gpu_flag == "b":
            location = peracommon.GpuLocation.MOTHERBOARD
        elif gpu_flag == "n":
            location = peracommon.GpuLocation.NONE
        else:
            location = None
    return location


def print_output(output: str, path: str):
    print("\nThe following output can be copy-pasted into the 'Bulk Add' page of the TARALLO, from '[' to ']':\n")
    print(output)

    with open(os.path.join(path, "copy_this_to_tarallo.json"), "w") as f:
        f.write(output)

    path = path.rstrip("/")
    print(
        f"You can also transfer the generated JSON file {path}/copy_this_to_tarallo.json to your PC with 'scp {path}/copy_this_to_tarallo.json <user>@<your_PC's_IP>:/path/on/your/PC' right from this terminal."
    )


def get_code() -> Optional[str]:
    if not args.code:
        code = input("Does this have a code already? (optional, ENTER to skip): ").strip()
    else:
        code = args.code
    if code and code != "":
        return code
    return None


def get_owner() -> Optional[str]:
    if not args.owner:
        owner = input("Do you want to add a owner? (optional, ENTER to skip): ").strip()
    else:
        owner = args.owner
    if owner and owner != "":
        return owner
    return None


def generate_parser():
    parser = argparse.ArgumentParser(
        description="Parse the files generated with generate_files.sh and " "get all the possible info out of them",
        epilog="If no argument is given, then this script will interactively guide you to run the PERACOTTA data gathering package."
        "Alternatively, you can choose to pass either the path to the directory where you want the files to be generated, the gpu location, or both."
        "In this case, the script will only become interactive when needed, and it won't ask you anything if you pass both the path and the gpu location.",
    )
    parser.add_argument(
        "-f",
        "--files",
        action="store_true",
        default=False,
        required=False,
        help="reuse previously generated files",
    )
    parser.add_argument(
        "--code",
        action="store",
        default=None,
        required=False,
        help="set the code assigned by T.A.R.A.L.L.O",
    )
    parser.add_argument("--owner", action="store", default=None, required=False, help="set a owner")
    parser.add_argument("-p", "--parsers", action="store", default=None, required=False, help="which parsers to use")
    gpu_group = parser.add_argument_group("GPU Location").add_mutually_exclusive_group(required=False)
    gpu_group.add_argument(
        "-g",
        "--gpu",
        action="store_true",
        default=False,
        help="computer has dedicated GPU",
    )
    gpu_group.add_argument(
        "-c",
        "--cpu",
        action="store_true",
        default=False,
        help="GPU is integrated inside the CPU",
    )
    gpu_group.add_argument(
        "-b",
        "--motherboard",
        action="store_true",
        default=False,
        help="GPU is integrated inside the motherboard",
    )
    gpu_group.add_argument(
        "--gpu-none",
        action="store_true",
        default=False,
        help="There's no GPU at all",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="print some warning messages",
    )
    parser.add_argument(
        "path",
        action="store",
        nargs="?",
        type=str,
        help="optional path where generated files are stored",
    )
    return parser


def main_():
    args = generate_parser().parse_args()

    try:
        main(args)
    except KeyboardInterrupt:
        print("\n[blue]Quitting...[/]")


if __name__ == "__main__":
    main_()



import json


def parse_win_chassis_specs(the_dir: str):
    with open(f"{the_dir}/chassis.win", "r") as file:
        data = json.load(file)
        object = [
            {
                "features": {
                    "brand": data["Manufacturer"],
                    "sn": data["SerialNumber"],
                },
                "type": "case",
            }
        ]
        return object


def parse_win_cpu_specs(the_dir: str):
    architectures = {
        0: "x86-32",
        1: "mips",
        2: "alpha",
        3: "powerpc",
        6: "ia64",
        9: "x86-64",
    }
    object = []
    with open(f"{the_dir}/lscpu.win", "r") as file:
        data = json.load(file)
        object.append(
            {
                "brand": data["Manufacturer"],
                "model": data["Name"],
                "features": {
                    "type": "cpu",
                    "isa": architectures[data["Architecture"]],
                    "core-n": data["NumberOfCores"],
                    "thread-n": data["ThreadCount"],
                    "frequency-hertz": int(data["MaxClockSpeed"]) * 1000000,
                },
                "type": "cpu",
            }
        )
    with open(f"{the_dir}/graphics.win", "r") as file:
        data = json.load(file)
        for entry in data:
            if "Service" in entry and entry["Service"] == "igfx":
                object[0]["features"]["integrated-graphics-brand"] = entry["Manufacturer"]
                object[0]["features"]["integrated-graphics-model"] = entry["Name"]
                break
    return object


def parse_win_ram_specs(the_dir: str):
    with open(f"{the_dir}/dimms.win", "r") as file:
        data = json.load(file)
        object = []
        for entry in data:
            object.append(
                {
                    "brand": entry["Manufacturer"],
                    "model": entry["PartNumber"],
                    "features": {
                        "frequency-hertz": entry["Speed"] * 1000000,
                        "capacity-byte": entry["Capacity"],
                        "ram-type": "",
                        "ram-ecc": "",
                        "ram-timings": "",
                        "sn": entry["SerialNumber"],
                    },
                    "type": "ram",
                }
            )
    return object


def parse_win_motherboard_specs(the_dir: str):
    with open(f"{the_dir}/baseboard.win", "r") as file:
        data = json.load(file)
        object = [
            {
                "brand": data["Manufacturer"],
                "model": data["Product"],
                "features": {
                    "parallel-ports-n": 0,
                    "usb-ports-n": 0,
                    "mini-jack-ports-n": 0,
                    "vga-ports-n": 0,
                    "serial-ports-n": 0,
                    "sata-ports-n": 0,
                    "ide-ports-n": 0,
                    "ps2-ports-n": 0,
                    "ethernet-ports-1000m-n": 0,
                },
                "type": "motherboard",
            }
        ]
    with open(f"{the_dir}/lspci.win", "r") as file:
        data = json.load(file)
        for entry in data:
            pnp_class = entry["PNPClass"]
            if pnp_class == "USB":
                object[0]["features"]["usb-ports-n"] += 1
                continue
            elif pnp_class == "USB":
                object[0]["features"]["usb-ports-n"] += 1
                continue
            elif pnp_class == "AudioEndpoint":
                object[0]["features"]["mini-jack-ports-n"] += 1
                continue
            elif pnp_class == "DiskDrive":
                object[0]["features"]["sata-ports-n"] += 1
                continue
    return object



import pytarallo.Errors
from PyQt6 import QtCore, QtGui, QtWidgets, uic
from pytarallo import Tarallo

from .constants import PATH


class Uploader(QtCore.QThread):
    successEvent = QtCore.pyqtSignal(str)
    failureEvent = QtCore.pyqtSignal(str, str)

    def __init__(
        self,
        data: list[dict],
        tarallo_url: str,
        tarallo_token: str,
        bulk_identifier: str,
        overwrite: bool,
    ):
        super().__init__()
        self.data = data
        self.tarallo_url = tarallo_url
        self.tarallo_token = tarallo_token
        self.bulk_identifier = bulk_identifier
        self.overwrite = overwrite

    def run(self) -> None:
        try:
            tarallo = Tarallo.Tarallo(self.tarallo_url, self.tarallo_token)
            ver = tarallo.bulk_add(self.data, self.bulk_identifier, self.overwrite)
            if ver:
                # TODO: use generated identifier if none was provided
                self.successEvent.emit(self.bulk_identifier)
            else:
                self.failureEvent.emit("cannot_upload", self.bulk_identifier)

        except pytarallo.Errors.NoInternetConnectionError:
            self.failureEvent.emit("cannot_reach", self.bulk_identifier)


class TaralloUploadDialog(QtWidgets.QDialog):
    signal = QtCore.pyqtSignal(bool, str, name="event")

    def __init__(self, parent: QtWidgets.QMainWindow, bulk_id: str = ""):
        super().__init__(parent)
        uic.loadUi(PATH["TARALLOUPLOADDIALOG"], self)

        self.setWindowTitle("Set bulk identifier")
        self.bulkLineEdit = self.findChild(QtWidgets.QLineEdit, "bulkLineEdit")
        self.bulkLineEdit.setText(bulk_id)
        self.okButton = self.findChild(QtWidgets.QPushButton, "okButton")
        self.okButton.clicked.connect(self.ok_signal)
        self.cancelButton = self.findChild(QtWidgets.QPushButton, "cancelButton")
        self.cancelButton.clicked.connect(self.cancel_signal)
        self.overwriteCheckBox = self.findChild(QtWidgets.QCheckBox, "overwriteCheckBox")
        self.show()

    def ok_signal(self):
        self.signal.emit(self.overwriteCheckBox.isChecked(), self.bulkLineEdit.text())
        self.close()

    def cancel_signal(self):
        self.close()


def tarallo_success_dialog(url: str):
    dialog: QtWidgets.QMessageBox = QtWidgets.QMessageBox(
        QtWidgets.QMessageBox.Icon.Information,
        "Upload successful",
        "Upload successful! Now go to TARALLO and finish the job.",
    )
    std_width = QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxWarning).availableSizes()[-1].width()
    dialog.setIconPixmap(QtGui.QPixmap(PATH["ICON"]).scaledToWidth(std_width, QtCore.Qt.TransformationMode.SmoothTransformation))
    dialog.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
    view_on_tarallo_button = dialog.addButton("View on TARALLO", dialog.ButtonRole.ActionRole)
    dialog.exec()
    if dialog.clickedButton() == view_on_tarallo_button:
        url = QtCore.QUrl(url)
        if not QtGui.QDesktopServices.openUrl(url):
            QtWidgets.QMessageBox.warning(dialog, "Cannot Open Url", f"Could not open url {url}")
        return True
    return False



import sys

from PyQt6 import QtCore, QtWidgets

from peracotta import commons
from peracotta.constants import PATH


class PeracottaThread(QtCore.QThread):
    updateEvent = QtCore.pyqtSignal(list, name="update")
    errorEvent = QtCore.pyqtSignal(str, str, name="error")
    startEvent = QtCore.pyqtSignal()

    def __init__(self, main_window: QtWidgets.QMainWindow):
        super().__init__()
        self.main_window = main_window

        self.gpu_location = None
        self.owner = ""
        self.files_path = PATH["TMP_FILES"]
        self.raw_files_path = ""
        self.generate_files = False
        self.filters = set()
        self.use_sudo = False
        self.sudo_passwd = None
        # self.set_default_values()

    def begin(self, generate_files: bool = True, raw_files_path: str = ""):
        self.generate_files = generate_files
        self.raw_files_path = raw_files_path
        self.start()

    def run(self) -> None:
        self.startEvent.emit()
        result = []
        try:
            if sys.platform == "win32":
                from scripts.get_windows_specs import generate_win_files

                generate_win_files()
                result = self.process_win_files()
            else:
                if self.generate_files:
                    # message = peracommon.check_required_files(self.files_path, is_gui=True)
                    # if message != "":
                    #     QtWidgets.QMessageBox.critical(self.main_window, "Error", message)
                    try:
                        self.files_path = commons.generate_files(self.files_path, self.use_sudo, self.sudo_passwd)
                    except commons.SudoError as error:
                        self.errorEvent.emit("Sudo error", str(error))
                        return
                    except commons.GenerateFilesError as error:
                        self.errorEvent.emit("Generate files error", str(error))
                        return
                    if self.files_path is None:
                        QtWidgets.QMessageBox.warning(self.main_window, "Critical error", "Failed to generate files")
                        return
                result = commons.call_parsers(
                    self.files_path,
                    set(self.filters),
                    self.gpu_location,
                    False,
                )
            if self.owner != "":
                result = commons.add_owner(result, self.owner)
            result = commons.split_products(result)
        finally:
            self.set_default_values()
        self.updateEvent.emit(result)

    def set_default_values(self):
        self.gpu_location = None
        self.owner = ""
        self.files_path = PATH["TMP_FILES"]
        self.raw_files_path = ""
        self.generate_files = False
        self.filters = set()
        self.use_sudo = False
        self.sudo_passwd = None

    def process_win_files(self):
        import parsers.windows_parser as win

        result = win.parse_win_cpu_specs(self.files_path)
        result = result + win.parse_win_chassis_specs(self.files_path)
        result = result + win.parse_win_ram_specs(self.files_path)
        result = result + win.parse_win_motherboard_specs(self.files_path)
        return result



#!/usr/bin/python3

# initial_chars_to_ignore is the length of the feature whose name the line begins with
# e.g. "Fundamental Memory Type" begins with 23 characters that are not all spaces, then n spaces to ignore,
# and finally there's the value needed, e.g. "DDR3 SDRAM"
from typing import List


def _ignore_spaces(line: str, initial_chars_to_ignore: int):
    relevant_part = line[initial_chars_to_ignore:]
    return relevant_part.strip()


def parse_decode_dimms(dimms: str, interactive: bool = False) -> List[dict]:
    # check based on output of decode-dimms v6250
    if "Number of SDRAM DIMMs detected and decoded: 0" in dimms or "Number of SDRAM DIMMs detected and decoded: " not in dimms:
        if interactive:
            print("decode-dimms was not able to find any RAM details")
        return []

    # split strings in 1 str array for each DIMM
    dimm_sections = dimms.split("Decoding EEPROM")
    # remove useless first part
    del dimm_sections[0]

    # create list of as many dimms as there are dimm_sections
    dimms = [
        {
            "type": "ram",
            "working": "yes",
        }
        for _ in range(len(dimm_sections))
    ]

    for i, dimm in enumerate(dimm_sections):
        manufacturer_data_type = None
        for line in dimm.splitlines():
            if line.startswith("Fundamental Memory type"):
                dimms[i]["ram-type"] = line.split(" ")[-2].lower()
                if dimms[i]["ram-type"] == "unknown":
                    del dimms[i]["ram-type"]

            if line.startswith("Maximum module speed"):
                freq = line.split(" ")[-3:-1]
                dimms[i]["frequency-hertz"] = int(freq[0])
                if "KHz" in freq[1] or "kHz" in freq[1]:
                    dimms[i]["frequency-hertz"] *= 1000
                elif "MHz" in freq[1]:
                    dimms[i]["frequency-hertz"] *= 1000 * 1000
                elif "GHz" in freq[1]:
                    dimms[i]["frequency-hertz"] *= 1000 * 1000 * 1000
                # The official thing is 667 MHz even if they run at 666 MHz
                if dimms[i]["frequency-hertz"] == 666000000:
                    dimms[i]["frequency-hertz"] = 667000000

            if line.startswith("Size"):
                cap = line.split(" ")[-2:]
                dimms[i]["capacity-byte"] = int(cap[0])
                if "KB" in cap[1] or "kB" in cap[1]:
                    dimms[i]["capacity-byte"] *= 1024
                elif "MB" in cap[1]:
                    dimms[i]["capacity-byte"] *= 1024 * 1024
                elif "GB" in cap[1]:
                    dimms[i]["capacity-byte"] *= 1024 * 1024 * 1024

            # alternatives to "Manufacturer" are "DRAM Manufacturer" and "Module Manufacturer"
            if "---=== Manufacturer Data ===---" in line:
                manufacturer_data_type = "DRAM Manufacturer"

            if "---=== Manufacturing Information ===---" in line:
                manufacturer_data_type = "Manufacturer"

            if manufacturer_data_type and line.startswith(manufacturer_data_type):
                dimms[i]["brand"] = _ignore_spaces(line, len(manufacturer_data_type))

            # This seems to always be the model (or at least never be the serial number)
            if line.startswith("Part Number"):
                model = _ignore_spaces(line, len("Part Number"))
                if model.lower() != "undefined":
                    dimms[i]["model"] = model

            # part number can be overwritten by serial number if present
            if line.startswith("Assembly Serial Number"):
                dimms[i]["sn"] = _ignore_spaces(line, len("Assembly Serial Number"))
                if dimms[i]["sn"].startswith("0x"):
                    try:
                        dimms[i]["sn"] = str(int(dimms[i]["sn"][2:], base=16))
                    except ValueError:
                        # Ooops, this isn't an hex number after all...
                        pass

            if line.startswith("Module Configuration Type"):
                if "Data Parity" in line or "Data ECC" in line or "Address/Command Parity" in line:
                    dimms[i]["ram-ecc"] = "yes"
                else:
                    dimms[i]["ram-ecc"] = "no"

            # Two (or more) spaces after because there are lines like "tCL-tRCD-tRP-tRAS as ..."
            if line.startswith("tCL-tRCD-tRP-tRAS  "):
                dimms[i]["ram-timings"] = _ignore_spaces(line, len("tCL-tRCD-tRP-tRAS"))

        if "ram-ecc" not in dimms[i] and len(dimms[i]) > 2:
            dimms[i]["ram-ecc"] = "no"

    return dimms


if __name__ == "__main__":
    import json
    import sys

    try:
        with open(sys.argv[1], "r") as f:
            input_file = f.read()
        print(json.dumps(parse_decode_dimms(input_file), indent=2))
    except BaseException as e:
        print(str(e))
        exit(1)



#!/usr/bin/env python3


"""
Read "lscpu" output
"""
from typing import List


def parse_lscpu(lscpu: str) -> List[dict]:
    cpu = {
        "type": "cpu",
        "working": "yes",
    }

    tmp_freq = None
    sockets = 1

    for line in lscpu.splitlines():
        if "Architecture:" in line:
            architecture = line.split("Architecture:")[1].strip()
            if architecture == "x86_64":
                cpu["isa"] = "x86-64"
            if architecture in ("i686", "i586", "i486", "i386"):
                cpu["isa"] = "x86-32"
        elif "CPU op-mode(s):" in line:
            architecture = line.split("CPU op-mode(s):")[1].strip()
            if cpu["isa"].startswith("x86"):
                if "64-bit" in architecture:
                    cpu["isa"] = "x86-64"
        elif "Model name:" in line:
            tmp = line.split("Model name:")[1].rsplit("@", 1)
            cpu["model"] = tmp[0].strip()
            if "@" in line:
                tmp_freq = tmp[1].replace("GHz", "").strip()
            elif line.endswith("GHz"):
                tmp_freq = line.rsplit(" ", 1)[1][:-3]

            if cpu["model"].startswith("Intel"):
                # To remove "(R)", or don't if it's not there
                cpu["model"] = cpu["model"].split(" ", 1)[1]
            if cpu["model"].endswith("-Core Processor"):
                cpu["model"] = cpu["model"].rsplit(" ", 2)[0]

            # Remove some more lapalissades and assorted tautologies
            cpu["model"] = (
                cpu["model"]
                .replace("(R)", " ")
                .replace("(TM)", " ")
                .replace("(tm)", " ")
                .replace("CPU", "")
                .replace("AMD", " ")
                .replace("Dual-Core", "")
                .replace("Quad-Core", "")
                .replace("Octa-Core", "")
                .replace("Processor", "")
                .replace("processor", "")
                .strip()
            )

            while "  " in cpu["model"]:
                cpu["model"] = cpu["model"].replace("  ", " ")

        elif "Vendor ID:" in line:
            cpu["brand"] = line.split("Vendor ID:")[1].strip()
            if cpu["brand"] == "GenuineIntel":
                cpu["brand"] = "Intel"
            elif cpu["brand"] == "AuthenticAMD":
                cpu["brand"] = "AMD"

        elif "CPU max MHz:" in line:
            # It's formatted with "%.4f" by lscpu, at the moment
            # https://github.com/karelzak/util-linux/blob/master/sys-utils/lscpu.c#L1246
            # .replace() needed because "ValueError: could not convert string to float: '3300,0000'"
            frequency_mhz = float(line.split("CPU max MHz:")[1].strip().replace(",", "."))
            cpu["frequency-hertz"] = int(frequency_mhz * 1000 * 1000)

        elif "CPU MHz:" in line and "frequency-hertz" not in cpu:
            # This may not exist anymore (?) but we should use it as a fallback
            frequency_mhz = float(line.split("CPU MHz:")[1].strip().replace(",", "."))
            cpu["frequency-hertz"] = int(frequency_mhz * 1000 * 1000)

        elif "Thread(s) per core:" in line:
            cpu["thread-n"] = int(line.split("Thread(s) per core:")[1].strip())

        elif "Core(s) per socket:" in line:
            cpu["core-n"] = int(line.split("Core(s) per socket:")[1].strip())
            if "thread-n" in cpu:
                cpu["thread-n"] *= cpu["core-n"]

        elif "Socket(s):" in line:
            sockets = int(line.split("Socket(s):")[1].strip())

    if tmp_freq is not None:
        cpu["frequency-hertz"] = int(float(tmp_freq.replace(",", ".")) * 1000 * 1000 * 1000)

    cpu = [cpu]
    if sockets > 1:
        cpu = cpu * sockets

    return cpu


if __name__ == "__main__":
    import json
    import sys

    try:
        with open(sys.argv[1], "r") as f:
            input_file = f.read()
        print(json.dumps(parse_lscpu(input_file), indent=2))
    except BaseException as e:
        print(str(e))
        exit(1)



import os

import importlib.resources

basedir = importlib.resources.files("peracotta")

logdir_path = basedir.joinpath("logs")

URL = {
    "website": "https://weeeopen.polito.it",
    "source_code": "https://github.com/WEEE-Open/peracotta",
}

VERSION = "2.2.4"

PATH = {
    "UI": "assets/interface.ui",
    "TARALLOUPLOADDIALOG": "assets/uploadTaralloDialog.ui",
    "ERRORDIALOG": "assets/error.ui",
    "JSON": "copy_this_to_tarallo.json",
    "FEATURES": "features.json",
    "THEMES": "assets/themes/",
    "TMP_FILES": "tmp/",
    "ICON": "assets/ui/pear_emoji.png",
}
for k, v in PATH.items():
    PATH[k] = os.path.join(basedir, v)

ICON = {
    "case": "assets/toolbox/case.png",
    "ram": "assets/toolbox/ram.png",
    "cpu": "assets/toolbox/cpu.png",
    "graphics-card": "assets/toolbox/gpu.png",
    "odd": "assets/toolbox/odd.png",
    "hdd": "assets/toolbox/hdd.png",
    "ssd": "assets/toolbox/ssd.png",
    "motherboard": "assets/toolbox/motherboard.png",
    "wifi-card": "assets/toolbox/wifi-card.png",
    "psu": "assets/toolbox/psu.png",
    "monitor": "assets/toolbox/monitor.png",
    "keyboard": "assets/toolbox/keyboard.png",
    "mouse": "assets/toolbox/mouse.png",
}

for k, v in ICON.items():
    ICON[k] = os.path.join(basedir, v)



def print_feature(feature, value, feature_type=None):
    unit = _name_to_unit(feature)
    if unit is None:
        if feature_type == "i":
            return int(value)
        elif feature_type == "d":
            return float(value)
        else:
            return str(value)
    else:
        return _print_value(unit, value)


def name_to_unit(name):
    return name.split("-")[-1]


def printable_to_value(unit, input_value):
    if not isinstance(input_value, str):
        return input_value
    string = input_value.strip()
    if len(string) <= 0:
        raise ValueError
    if unit == "n":
        return int(input_value)

    i = 0
    pure = False
    for i in range(0, len(input_value)):
        if not (input_value[i].isdigit() or (input_value[i] in (".", ","))):
            break
    else:
        pure = True

    if pure:
        number = float(input_value)
    else:
        if i == 0:
            raise ValueError
        number = float(input_value[0:i])

    exp = 0
    if unit == "mm":
        exp = 0
    elif pure:
        exp = 0
    else:
        for char in input_value[i:]:
            char = char.lower()
            if char.isalpha():
                exp = _prefix_to_exponent(char)
                break
    if unit == "byte":
        base = 1024
    else:
        base = 1000

    return number * (pow(base, exp))


def _prefix_to_exponent(char):
    if char == "k":
        return 1
    if char == "m":
        return 2
    if char == "g":
        return 3
    if char == "t":
        return 4
    if char == "p":
        return 5
    if char == "e":
        return 6
    return 0


def _name_to_unit(name):
    pieces = name.split("-")
    units = {
        "byte": "byte",
        "hertz": "Hz",
        "decibyte": "B",
        "ampere": "A",
        "volt": "V",
        "watt": "W",
        "rpm": "rpm",
        "mm": "mm",
        "inch": "in.",
        "gram": "g",
    }
    if pieces[-1] in units:
        return units[pieces[-1]]
    else:
        return None


def _print_value(unit, value):
    if unit == "n":
        return str(value)
    elif unit == "rpm" or unit == "mm" or unit == "in.":
        return f"{_format_value(value)} {unit}"
    elif unit == "byte":
        return _append_unit(value, "B", 1024)
    else:
        return _append_unit(value, unit, 1000)


def _append_unit(value, unit, base_unit=1000):
    prefix = 0
    while value >= base_unit and prefix <= 6:
        value /= base_unit
        prefix += 1
    i = ""
    if prefix > 0 and base_unit == 1024:
        i = "i"
    return f"{_format_value(value)} {_prefix_to_printable(prefix, base_unit == 1024)}{i}{unit}"


def _format_value(value: float):
    return f"{value:g}"


def _prefix_to_printable(places, big_k=False):
    prefixes = {
        0: "",
        1: "K" if big_k else "k",
        2: "M",
        3: "G",
        4: "T",
        5: "P",
        6: "E",
        # -1: 'm',
        # -2: 'µ',
        # -3: 'n',
    }
    return prefixes[places]



#!/usr/bin/env python3
from typing import List, Optional

connectors_map = {
    "PS/2": "ps2-ports-n",
    "Access Bus (USB)": "usb-ports-n",
    "DB-25 male": "parallel-ports-n",
    "DB-25 female": "parallel-ports-n",
    "DB-9 male": "serial-ports-n",
    "DB-15 female": "vga-ports-n",
    "Mini Jack (headphones)": "mini-jack-ports-n",
    "RJ-45": "ethernet-ports-n",  # not a real feature in T.A.R.A.L.L.O., since it's not yet known if it's 100 or 1000
    "Mini DisplayPort": "mini-displayport-ports-n",
    "Thunderbolt": "thunderbolt-ports-n",
    "HDMI": "hdmi-ports-n",
    "SATA0": "sata-ports-n",
    "SATA1": "sata-ports-n",
    "SATA2": "sata-ports-n",
    "SATA3": "sata-ports-n",
    "SATA4": "sata-ports-n",
    "SATA5": "sata-ports-n",
    "SATA6": "sata-ports-n",
    "SATA7": "sata-ports-n",
    "USB0": "usb-ports-n",
    "USB1": "usb-ports-n",
    "USB2": "usb-ports-n",
    "USB3": "usb-ports-n",
    "USB4": "usb-ports-n",
    "USB5": "usb-ports-n",
    "USB6": "usb-ports-n",
    "USB7": "usb-ports-n",
    "USB8": "usb-ports-n",
    "LINE_IN": "mini-jack-ports-n",
    "IEEE 1394": "firewire-ports-n",
    "RJ-11": "rj11-ports-n",
    "On Board Sound Input From CD-ROM": None,
    "On Board Floppy": None,
    "CHASSIS REAR FAN": None,
    "CHASSIS FAN": None,
    "CPU FAN": None,
    "FNT USB": None,
    "FP AUD": None,
    "ATX_PWR": None,
    "9 Pin Dual Inline (pin 10 cut)": None,  # Internal USB header?
    "Microphone": None,  # Internal microphone, not a connector
    "Speaker": None,
    "SPEAKER (SPKR)": None,  # Internal speaker (header)
    "FP_AUDIO": None,
    "PASSWORD CLEAR (PSWD)": None,
    "HOOD LOCK (HLCK)": None,
    "HOOD SENSE (HSENSE)": None,
    "TPM SECURITY (SEC)": None,
}
connectors_map_tuples = {
    ("On Board IDE", None, "*IDE*", None): "ide-ports-n",
    ("On Board IDE", None, None, "* IDE"): "ide-ports-n",
    ("On Board IDE", None, "PRIMARY*", None): "ide-ports-n",
    ("On Board IDE", None, "SECONDARY*", None): "ide-ports-n",
    (None, None, "*ESATA*", None): "esata-ports-n",
    ("On Board IDE", None, "SATA*", None): "sata-ports-n",
    (
        "On Board IDE",
        None,
        "*SATA*",
        None,
    ): "sata-ports-n",  # Don't add *SATA, it matches ESATA...
    (None, None, None, "REAR LINE IN"): "mini-jack-ports-n",
    (None, None, None, "REAR HEADPHONE/LINEOUT"): "mini-jack-ports-n",
    (None, None, "*FAN", None): None,
    (None, None, "CHA_FAN*", None): None,
    (None, None, "FRNT AUD*", None): None,  # Front audio is not part of the motherboard
    ("SAS/SATA Plug Receptacle", None, "SATA*", None): "sata-ports-n",
    ("SAS/SATA Plug Receptacle", None, "*EIDE", None): "ide-ports-n",
    ("SAS/SATA Plug Receptacle", None, "SAS*", None): "sas-sata-ports-n",
}
extra_connectors = {
    "MagSafe DC Power": {"power-connector": "proprietary"},
}


def parse_motherboard(baseboard_file: str, connectors_file: str, net_file: str, interactive: bool = False) -> List[dict]:
    baseboard = _get_baseboard(baseboard_file)
    baseboard = _get_connectors(connectors_file, baseboard, interactive)
    baseboards = _get_net(net_file, baseboard, interactive)
    return baseboards


def _get_baseboard(baseboard: str) -> dict:
    mobo = {
        "type": "motherboard",
        "working": "yes",
    }

    for line in baseboard.splitlines():
        if "Manufacturer:" in line:
            mobo["brand"] = line.split("Manufacturer:")[1].strip()

        elif "Product Name:" in line:
            mobo["model"] = line.split("Product Name:")[1].strip()

        elif "Serial Number:" in line:
            mobo["sn"] = line.split("Serial Number:")[1].strip().strip(".")

    # Get MSI internal codename
    if mobo.get("brand", "").upper().startswith("MICRO-STAR INTERNATIONAL"):
        model = mobo.get("model", "")
        model_parts = model.split("(")
        if len(model_parts) == 2:
            mobo["model"] = model_parts[0].strip()
            mobo["internal-name"] = model_parts[1].rstrip(")").strip()

    return mobo


def _get_connectors(connectors_file: str, baseboard: dict, interactive: bool = False) -> dict:
    possible_connectors = set(connectors_map.values()) | set(connectors_map_tuples.values())
    possible_connectors.remove(None)
    connectors = dict(zip(possible_connectors, [0] * len(connectors_map)))

    # TODO: this part (is it needed?)
    # port_types = []
    # devices = output.split("On Board Device")
    # for device in devices:
    # 	type = device.split("Description:")
    # 	if len(type) > 1:
    # 		port_types.append(type[1].replace("\n", "").replace(" ", ""))

    warnings = []
    for section in connectors_file.split("\n\n"):
        if not section.startswith("Handle "):
            continue
        internal = get_dmidecoded_value(section, "Internal Connector Type:")
        external = get_dmidecoded_value(section, "External Connector Type:")
        internal_des = get_dmidecoded_value(section, "Internal Reference Designator:")
        external_des = get_dmidecoded_value(section, "External Reference Designator:")
        if external in ("None", "Other", "Not Specified"):
            if internal in ("None", "Other", "Not Specified"):
                if external_des in ("None", "Other", "Not Specified"):
                    connector = internal_des
                else:
                    connector = external_des
            else:
                connector = internal
        else:
            connector = external

        if connector in connectors_map:
            if connectors_map[connector] is not None:
                connectors[connectors_map[connector]] += 1
        elif connector in extra_connectors:
            # Dark magic: https://stackoverflow.com/a/26853961
            connectors = {**connectors, **(extra_connectors[connector])}
        else:
            found = find_connector_from_tuple(connectors, external, external_des, internal, internal_des)
            if not found:
                warning = f"Unknown connector: {internal} / {external} ({internal_des} / {external_des})"
                if interactive:
                    print(warning)
                warnings.append(warning)

    connectors_clean = {}
    # Keys to avoid changing dict size at runtime (raises an exception)
    for connector in connectors:
        if isinstance(connectors[connector], int):
            if connectors[connector] > 0:
                connectors_clean[connector] = connectors[connector]
        else:
            connectors_clean[connector] = connectors[connector]

    if len(warnings) > 0:
        warnings = "\n".join(warnings)
        # Dark magic: https://stackoverflow.com/a/26853961
        return {**baseboard, **connectors_clean, **{"notes": warnings}}
    else:
        # Somewhat less dark magic
        return {**baseboard, **connectors_clean}


def _get_net(net: str, baseboard: dict, interactive: bool = False) -> List[dict]:
    mergeit = {
        "ethernet-ports-100m-n": 0,
        "ethernet-ports-1000m-n": 0,
        "mac": [],
    }
    other_devices = []

    for line in net.split("\n"):
        if "u" in line:
            # USB adapters, ignore them
            continue
        line = line.split(" ", 3)
        if line[0].startswith("en"):
            if interactive and len(line) <= 2:
                print(f"Warning: cannot detect speed for Ethernet port {line[0]}, is it unconnected?")
            elif line[2] == "1000":
                mergeit["ethernet-ports-1000m-n"] += 1
            elif line[2] == "100":
                mergeit["ethernet-ports-100m-n"] += 1
            elif interactive:
                print(f"Warning: unknown speed for Ethernet port {line[0]}: {line[2]}")
            mergeit["mac"].append(line[1])
        if line[0].startswith("wl"):
            other_devices.append({"type": "wifi-card", "mac": line[1], "notes": f"Device name {line[0]}"})

    mergeit["mac"] = ", ".join(mergeit["mac"])

    if "ethernet-ports-n" in baseboard:
        found_ports = mergeit["ethernet-ports-100m-n"] + mergeit["ethernet-ports-1000m-n"]
        baseboard["ethernet-ports-n"] -= found_ports
        if baseboard["ethernet-ports-n"] > 0:
            if baseboard["ethernet-ports-n"] > 1:
                message = f"\nBIOS reported {baseboard['ethernet-ports-n']} more ethernet ports that were not found by the kernel"
            else:
                message = f"\nBIOS reported {baseboard['ethernet-ports-n']} more ethernet port that was not found by the kernel"
            if "notes" in baseboard:
                baseboard["notes"] += message
                baseboard["notes"] = baseboard["notes"].strip()
            else:
                baseboard["notes"] = message.strip()
        del baseboard["ethernet-ports-n"]

    if mergeit["ethernet-ports-100m-n"] <= 0:
        del mergeit["ethernet-ports-100m-n"]
    if mergeit["ethernet-ports-1000m-n"] <= 0:
        del mergeit["ethernet-ports-1000m-n"]
    if len(mergeit["mac"]) <= 0:
        del mergeit["mac"]
    baseboard = {**baseboard, **mergeit}

    # It even says the type is the exact same one it expects, however it's an error
    # noinspection PyTypeChecker
    return [baseboard] + other_devices


def find_connector_from_tuple(connectors, external, external_des, internal, internal_des):
    equal = False
    for tup in connectors_map_tuples:
        zipped = list(zip(tup, (internal, external, internal_des, external_des)))
        equal = True
        for mask, garbage_from_manufacturer in zipped:
            if mask is None:
                continue
            if mask.startswith("*") and mask.endswith("*"):
                if mask[1:-1] not in garbage_from_manufacturer:
                    equal = False
                    break
            elif mask.endswith("*"):
                if not garbage_from_manufacturer.startswith(mask[:-1]):
                    equal = False
                    break
            elif mask.startswith("*"):
                if not garbage_from_manufacturer.endswith(mask[1:]):
                    equal = False
                    break
            elif mask != garbage_from_manufacturer:
                equal = False
                break
        if equal:
            if connectors_map_tuples[tup] is not None:
                connectors[connectors_map_tuples[tup]] += 1
            return equal
    return equal


def get_dmidecoded_value(section: str, key: str) -> str:
    return section.split(key, 1)[1].split("\n", 1)[0].strip()


def parse_psu(chassis: Optional[dict]):
    if chassis.get("motherboard-form-factor") == "proprietary-laptop":
        return [{"type": "external-psu", "check": "missing-data", "working": "yes"}]
    else:
        return [{"type": "psu", "check": "missing-data", "working": "yes"}]


def parse_case(chassis_file: str, mobo: Optional[dict] = None) -> List[dict]:
    chassis = {"type": "case"}

    for line in chassis_file.splitlines():
        if "Manufacturer" in line:
            manufacturer = line.split("Manufacturer:")[1].strip()
            if len(manufacturer) > 0:
                chassis["brand"] = manufacturer

        # This is Desktop, Laptop, etc...
        elif "Type: " in line:
            ff = line.split("Type: ")[1].strip()
            if ff == "Laptop" or ff == "Notebook":  # Both exist in the wild and in tests, difference unknown
                chassis["motherboard-form-factor"] = "proprietary-laptop"
                if mobo and "motherboard-form-factor" not in mobo:
                    mobo["motherboard-form-factor"] = "proprietary-laptop"

        elif "Serial Number" in line:
            sn = line.split("Serial Number:")[1].strip().strip(".")
            if len(sn.strip("0")) <= 0:
                sn = ""
            if len(sn) > 0:
                chassis["sn"] = sn

    for key, value in chassis.items():
        if value == "Unknown":
            # Remove pointless values
            del chassis[key]

    return [chassis]


if __name__ == "__main__":
    import argparse
    import json

    parser = argparse.ArgumentParser(description="Parse dmidecode output")
    parser.add_argument("-b", "--baseboard", type=str, help="Path to baseboard.txt")
    parser.add_argument("-c", "--chassis", type=str, help="Path to chassis.txt")
    parser.add_argument("-p", "--ports", type=str, help="Path to connector.txt (ports)")
    parser.add_argument("-n", "--net", type=str, help="Path to net.txt")
    args = parser.parse_args()

    if args.ports is not None and args.baseboard is None:
        print("Provide a baseboard.txt file to detect connectors")
        exit(2)
    if args.net is not None and args.baseboard is None:
        print("Provide a baseboard.txt file to detect network cards")
        exit(2)
    if args.baseboard is not None and args.ports is None:
        print("Provide a connector.txt file to parse motherboard")
        exit(2)
    if args.baseboard is not None and args.net is None:
        print("Provide a net.txt file to parse motherboard")
        exit(2)

    try:
        if args.baseboard is not None:
            with open(args.baseboard, "r") as f:
                input_baseboard = f.read()
            with open(args.ports, "r") as f:
                input_ports = f.read()
            with open(args.ports, "r") as f:
                input_net = f.read()
            bb = parse_motherboard(input_baseboard, input_ports, input_net, True)
            print(json.dumps(bb, indent=2))
        if args.chassis is not None:
            with open(args.chassis, "r") as f:
                input_chassis = f.read()
            try:
                # This pointless check triggers NameError if bb is not defined
                # noinspection PyUnboundLocalVariable
                if bb is None:
                    pass
            except NameError:
                bb = None

            # bb is bery well defined, Pycharm doesn't believe it though
            # noinspection PyUnboundLocalVariable
            print(json.dumps(parse_case(input_chassis, bb), indent=2))

    except FileNotFoundError as e:
        print(str(e))
        exit(1)



import os

POWERSHELL = r"powershell -Command"


def get_win_cpu_specs() -> str:
    command = 'Get-CimInstance -ClassName Win32_Processor | Select-Object -ExcludeProperty "CIM*" | ConvertTo-Json'
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_ram_specs() -> str:
    command = "Get-WmiObject Win32_PhysicalMemory | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_motherboard_specs() -> str:
    command = "Get-WmiObject Win32_BaseBoard | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_chassis_specs() -> str:
    command = "Get-WmiObject Win32_SystemEnclosure | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_disks_specs() -> str:
    command = "Get-Disk | Select * | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_pnp_specs() -> str:
    command = "Get-PnpDevice | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_network_specs() -> str:
    command = "Get-NetAdapter | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def get_win_graphics_card_specs() -> str:
    command = "Get-WmiObject Win32_VideoController | ConvertTo-Json"
    data = os.popen(f'{POWERSHELL} "& {command}"').read()
    return data


def generate_win_files():
    os.makedirs("tmp", exist_ok=True)
    with open("tmp/lscpu.win", "w") as file:
        file.write(f"{get_win_cpu_specs()}")
        file.flush()
        file.close()
    with open("tmp/dimms.win", "w") as file:
        file.write(f"{get_win_ram_specs()}")
        file.flush()
        file.close()
    with open("tmp/baseboard.win", "w") as file:
        file.write(f"{get_win_motherboard_specs()}")
        file.flush()
        file.close()
    with open("tmp/chassis.win", "w") as file:
        file.write(f"{get_win_chassis_specs()}")
        file.flush()
        file.close()
    with open("tmp/disks.win", "w") as file:
        file.write(f"{get_win_disks_specs()}")
        file.flush()
        file.close()
    with open("tmp/lspci.win", "w") as file:
        file.write(f"{get_win_pnp_specs()}")
        file.flush()
        file.close()
    with open("tmp/net.win", "w") as file:
        file.write(f"{get_win_pnp_specs()}")
        file.flush()
        file.close()
    with open("tmp/graphics.win", "w") as file:
        file.write(f"{get_win_pnp_specs()}")
        file.flush()
        file.close()


def main():
    generate_win_files()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Read "lspci -v" and "glxinfo" outputs
"""

import re
from typing import List


def _read_lspci_output(gpu: dict, lspci_file: str, interactive: bool = False):
    lspci_sections = lspci_file.split("\n\n")

    for section in lspci_sections:
        if "VGA compatible controller:" in section:
            # removes "VGA compatible controller:"
            first_line = section.splitlines()[0].split(": ", 1)[1].strip()
            second_line = section.splitlines()[1].strip()
            part_between_square_brackets = None
            try:
                # take the first string between [] from the first line
                part_between_square_brackets = first_line.split("[")[1].split("]")[0]
            except IndexError:
                # there may not be an argument in between []
                pass

            if "Subsystem:" in second_line:
                # The model or model family is often repeated here, but removing it automatically is complicated
                gpu["brand"] = second_line.split("Subsystem: ")[1].split("[", 1)[0].strip()
                gpu["brand"] = gpu["brand"].replace("Integrated Graphics Controller", "").strip()

            # -----------------------------------------------------------------
            # AMD/ATI
            # -----------------------------------------------------------------
            if part_between_square_brackets is not None and ("AMD" in part_between_square_brackets or "ATI" in part_between_square_brackets):
                gpu["brand-manufacturer"] = part_between_square_brackets
                # take second string between []
                gpu["model"] = first_line.split("[")[2].split("]")[0]
                if "controller" in gpu["model"]:
                    gpu["model"] = section.splitlines()[1].split(" ")[-1]

            # -----------------------------------------------------------------
            # Nvidia
            # -----------------------------------------------------------------
            elif "NVIDIA" in first_line.upper():
                gpu["brand-manufacturer"] = "Nvidia"
                gpu["model"] = part_between_square_brackets
                if "brand" in gpu:
                    pieces = gpu["brand"].rsplit(" ", 1)
                    gpu["brand"] = pieces[0]
                    gpu["internal-name"] = pieces[1]

            # -----------------------------------------------------------------
            # Intel
            # -----------------------------------------------------------------
            elif "INTEL" in first_line.upper():
                gpu["brand-manufacturer"] = "Intel"
                if "Integrated Graphics" in first_line:
                    tmp_model = first_line.split("Intel Corporation ")[1].split(" Integrated Graphics")[0]
                    # if there are no numbers, e.g. "Core Processor", tmp_model is not a model number
                    if not re.search("\\d+", tmp_model):
                        tmp_model = ""
                elif "HD Graphics" in first_line:
                    tmp_model = first_line.split("Intel Corporation ")[1].split("(", 1)[0].strip()
                elif "[" in first_line and "]" in first_line:
                    tmp_model = first_line.split("[")[1].split("]")[0]
                else:
                    tmp_model = ""

                if tmp_model != "":
                    gpu["model"] = tmp_model

            # -----------------------------------------------------------------
            # VIA
            # -----------------------------------------------------------------
            elif first_line.startswith("VIA"):
                gpu["brand-manufacturer"] = "VIA"
                gpu["model"] = part_between_square_brackets

                tmp_model = first_line.split("[")[0]
                i = 0
                for i, char in enumerate("VIA Technologies, Inc. "):
                    if tmp_model[i] != char:
                        break
                gpu["internal-name"] = tmp_model[i:].strip()

            # -----------------------------------------------------------------
            # SiS
            # -----------------------------------------------------------------
            elif part_between_square_brackets == "SiS":
                # May be written somewhere else on other models, but we have so few SiS cards that it's difficult to
                # find more examples. Also, they haven't made any video card in the last 15 years or so.
                gpu["brand-manufacturer"] = part_between_square_brackets
                if "brand" in gpu and gpu["brand"].lower() == "silicon integrated systems":
                    gpu["brand"] = "SiS"
                gpu["model"] = first_line.split("]", 1)[1]
                # These may be useful for non-integrated cards, however the example ones are all integrated
                if " PCIE" in gpu["model"]:
                    gpu["model"] = gpu["model"].split(" PCIE", 1)[0].strip()
                elif " PCI/AGP" in gpu["model"]:
                    gpu["model"] = gpu["model"].split(" PCI/AGP", 1)[0].strip()
                if gpu["model"] in gpu["brand"]:
                    gpu["brand"] = gpu["brand"].split(gpu["model"], 1)[0].strip()
            else:
                gpu["brand-manufacturer"] = None
                if interactive:
                    print(
                        "I couldn't find the Video Card brand. The model was set to 'None' and is to be edited "
                        "logging into the TARALLO afterwards. The information you're looking for should be in the "
                        f"following 2 lines:\n{first_line}\n{second_line}\n"
                    )

            if gpu.get("model") and gpu.get("brand"):
                # Try to remove duplicate information
                gpu["brand"] = gpu["brand"].replace(gpu["model"], "").strip()
            else:
                if interactive:
                    print(
                        "I couldn't find the Integrated Graphics model. The model was set to 'None' and is to be "
                        "edited logging into the TARALLO afterwards. The information you're looking for should be in "
                        f"the following 2 lines:\n{first_line}\n{second_line}\n"
                    )
            break

    if gpu.get("internal-name"):
        # Same
        gpu["brand"] = gpu["brand"].replace(gpu["internal-name"], "").strip()

    if gpu.get("brand") == "":
        del gpu["brand"]

    if gpu.get("brand-manufacturer") and gpu.get("brand"):
        if gpu["brand-manufacturer"].lower() == gpu["brand"].lower():
            del gpu["brand-manufacturer"]

    if interactive:
        if "brand" not in gpu and "brand-manufacturer" not in gpu:
            print(
                "I couldn't find the Video Card brand. The model was set to 'None' and is to be edited logging "
                "into the TARALLO afterwards. The information you're looking for should be in the following 2 lines:"
            )
        if "capacity-byte" not in gpu:
            print(
                "A dedicated video memory couldn't be found. A generic video memory capacity was found instead, which "
                "could be near the actual value. Please humans, fix this error by hand."
            )


def _read_glxinfo_output(gpu: dict, glxinfo_file: str):
    for line in glxinfo_file.splitlines():
        # this line comes before the "Dedicated video memory" line
        # this basically saves a default value if the dedicated memory line cannot be found
        if "Video memory" in line:
            try:
                tmp_vid_mem = int(line.split(" ")[6].split(" ")[0][:-2])
                tmp_vid_mem_multiplier = line[-2:]
            except ValueError:
                continue

            _parse_capacity(gpu, tmp_vid_mem, tmp_vid_mem_multiplier)
            break

        if "Dedicated video memory" in line:
            try:
                tmp_vram = int(line.split(" ")[7].split(" ")[0])
                tmp_vram_multiplier = line[-2:]
            except ValueError:
                continue

            _parse_capacity(gpu, tmp_vram, tmp_vram_multiplier)
            break

    if not gpu.get("capacity-byte"):
        if "notes" in gpu:
            gpu["notes"] += "\n"
        else:
            gpu["notes"] = ""
        gpu["notes"] += "Could not find dedicated video memory, check glxinfo output"


def _parse_capacity(gpu, tmp_vram, tmp_vram_multiplier):
    capacity = _convert_video_memory_size(tmp_vram, tmp_vram_multiplier)
    # Round to the next power of 2
    # this may be different from human readable capacity...
    rounded = 2 ** (capacity - 1).bit_length()
    one_and_half = int(rounded / 2 * 1.5)
    # Accounts for 3 GB VRAM cards and similar
    # Yes they do exist, try to remove this part and watch tests fail (and the card was manually verified to be 3 GB)
    if one_and_half >= capacity:
        gpu["capacity-byte"] = one_and_half
    else:
        gpu["capacity-byte"] = rounded


def _convert_video_memory_size(capacity, units_of_measure):
    if units_of_measure == "GB":
        capacity *= 1024 * 1024 * 1024
    elif units_of_measure == "MB":
        capacity *= 1024 * 1024
    elif units_of_measure.upper() == "KB":
        capacity *= 1024
    else:
        capacity = -1

    return capacity


def parse_lspci_and_glxinfo(has_dedicated: bool, lspci_file: str, glxinfo_file: str, interactive: bool = False) -> List[dict]:
    gpu = {
        "type": "graphics-card",
        "working": "yes",
    }
    if has_dedicated:
        _read_lspci_output(gpu, lspci_file, interactive)
        _read_glxinfo_output(gpu, glxinfo_file)
    else:
        # integrated in mobo or cpu
        _read_lspci_output(gpu, lspci_file, interactive)
        # don't parse glxinfo because the VRAM is part of the RAM and varies
        if "capacity-byte" in gpu:
            del gpu["capacity-byte"]

    return [gpu]


if __name__ == "__main__":
    import argparse
    import json

    parser = argparse.ArgumentParser(description="Parse lspci/glxinfo output")
    parser.add_argument("lspci", type=str, nargs=1, help="path to lspci output")
    parser.add_argument("glxinfo", type=str, nargs=1, help="path to glxinfo output")
    parser.add_argument(
        "-d",
        "--dedicated",
        action="store_true",
        default=False,
        help="computer has dedicated GPU",
    )
    args = parser.parse_args()

    try:
        with open(args.lspci[0], "r") as f:
            input_lspci = f.read()
        with open(args.glxinfo[0], "r") as f:
            input_glxinfo = f.read()

        print(
            json.dumps(
                parse_lspci_and_glxinfo(args.dedicated, input_lspci, input_glxinfo),
                indent=2,
            )
        )
    except FileNotFoundError as e:
        print(str(e))
        exit(1)



from collections import defaultdict
from typing import Optional

from PyQt6 import QtCore, QtGui, QtWidgets

from peracotta.commons import item_only_features
from peracotta.constants import ICON

from . import prettyprinter


class ItemEnumDelegate(QtWidgets.QStyledItemDelegate):
    # class ItemEnumAlignDelegate(QtWidgets.QStyledItemDelegate):
    #     def initStyleOption(self, option, index):
    #         super().initStyleOption(option, index)
    #         option.displayAlignment = QtCore.Qt.AlignLeft

    def createEditor(self, parent, option, index):
        the_type = str(index.model().data(index, QtCore.Qt.ItemDataRole.UserRole))
        if the_type == "e":
            editor = QtWidgets.QComboBox(parent)
            editor.currentTextChanged.connect(self.handle_editor_change)
            editor.setEditable(True)
            return editor
        else:
            return super().createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        model: CustomTableModel = index.model()
        if isinstance(editor, QtWidgets.QComboBox):
            values = model.row_all_enum_values_for_editor(index.row())
            current = model.row_current_enum_value_for_editor(index.row())
            if values and current:
                # editor.setItemDelegate(self.ItemEnumAlignDelegate(editor))
                for k, v in values.items():
                    editor.addItem(v, k)
                    i = editor.count() - 1
                    if current == k:
                        editor.setCurrentIndex(i)
        else:
            return super().setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        if isinstance(editor, QtWidgets.QComboBox):
            model.setData(index, editor.currentData(), QtCore.Qt.EditRole)
        else:
            return super().setModelData(editor, model, index)

    def handle_editor_change(self):
        editor = self.sender()
        self.commitData.emit(editor)
        self.closeEditor.emit(editor)


class CustomTableView(QtWidgets.QTableView):
    def __init__(self):
        super().__init__()
        self.setItemDelegateForColumn(1, ItemEnumDelegate())

    def minimumSizeHint(self) -> QtCore.QSize:
        default_size = super().minimumSizeHint()

        frame = self.frameWidth() * 2

        header = self.verticalHeader().sizeHint().height()
        rows = self.verticalHeader().length()
        # rows = self.model().rowCount() * self.horizontalHeader().defaultSectionSize()
        h = header + rows + frame
        # print(f"{header} + {rows} + {frame} = {h} > {default_size.height()}")

        return QtCore.QSize(default_size.width(), h)

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        menu = QtWidgets.QMenu(self)
        remove_action = QtGui.QAction("Remove feature", self)
        remove_action.triggered.connect(self.remove_row)
        menu.addAction(remove_action)
        menu.popup(QtGui.QCursor.pos())

    def remove_row(self):
        self.model().removeRow(self.selectedIndexes()[1].row(), self.selectedIndexes()[1].parent())


class CustomTableModel(QtCore.QAbstractTableModel):
    emergency_resize = QtCore.pyqtSignal(name="emergency_resize")

    def __init__(self, data: list[dict], item_features: dict, product: Optional[dict], default_features: dict):
        super().__init__()

        self._data = data
        self.ref_features = item_features
        self.ref_product = product
        self.default_features = default_features

        self.features = {}
        self.feature_keys = []
        self._productize(item_features, product)

    def _productize(self, item_features: dict, product: Optional[dict]):
        if product:
            self.features = product["features"].copy()
        else:
            self.features = {}
        self.features.update(item_features)
        self.feature_keys = list(self.features)

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.feature_keys)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 2

    # noinspection PyMethodOverriding
    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole and orientation == QtCore.Qt.Orientation.Horizontal:
            if section == 0:
                return "Feature"
            else:
                return "Value"
        return None

    def flags(self, index):
        if index.column() == 1:
            return QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def row_all_enum_values_for_editor(self, row: int) -> Optional[dict[str, str]]:
        if row < 0 or row >= len(self.feature_keys):
            return None

        name = self.feature_keys[row]
        feature_type = self._get_feature_type(name)
        if feature_type == "e":
            return self.default_features["values"][name]
        return None

    def row_current_enum_value_for_editor(self, row: int):
        if row < 0 or row >= len(self.features):
            return None

        name = self.feature_keys[row]
        return self.features.get(name)

    # def _row_to_name(self, row) -> Optional[str]:
    #     for i, name in enumerate(self.combined):
    #         if i == row:
    #             return name
    #         else:
    #             return None

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        row = index.row()
        if row < 0 or row >= len(self.feature_keys):
            return None

        if role == QtCore.Qt.ItemDataRole.DisplayRole or role == QtCore.Qt.ItemDataRole.EditRole or role == QtCore.Qt.ItemDataRole.UserRole:
            column = index.column()
            name = self.feature_keys[row]
            if column == 0:
                return self.default_features["names"].get(name, name)
            elif column == 1:
                feature_type = self._get_feature_type(name)
                if role == QtCore.Qt.ItemDataRole.UserRole:
                    return feature_type
                value = self.features[name]
                if feature_type == "e":
                    return self.default_features["values"][name].get(value, value)
                elif feature_type in ("d", "i"):
                    return prettyprinter.print_feature(name, value, feature_type)
                else:
                    return value
        elif role == QtCore.Qt.ItemDataRole.TextAlignmentRole:
            column = index.column()
            if column == 0:
                return QtCore.Qt.AlignmentFlag.AlignLeft + QtCore.Qt.AlignmentFlag.AlignVCenter
            elif column == 1:
                return QtCore.Qt.AlignmentFlag.AlignLeft + QtCore.Qt.AlignmentFlag.AlignVCenter

        return None

    # noinspection PyMethodOverriding
    def setData(self, index, value, role):
        if role == QtCore.Qt.ItemDataRole.EditRole:
            row = index.row()
            col = index.column()
            if col != 1:
                return False
            if row < 0 or row >= len(self.feature_keys):
                return False

            # Feature name
            name = self.feature_keys[row]
            # Normalize and validate, with EXTREME intensity
            ok, value = self.extreme_validation(name, value)
            if ok:
                # Add to features, this is a local copy of merged
                # item and product features
                self.features[name] = value
                product_to_add = None
                # Search if a product is there but not linked,
                # this is needed when a new item is added
                # since brand, model and variant are filled
                # one at a time
                if not self.ref_product:
                    product_to_add = ToolBoxWidget.find_matching_product(self._data, self.features)
                    if product_to_add:
                        self.beginResetModel()
                        self._productize(self.features, product_to_add)
                # If this feature exists in the product, add it there.
                # Otherwise, if the item has it, add it there.
                # Otherwise, find where it should be added.
                if self.ref_product and name in self.ref_product["features"]:
                    self.ref_product["features"][name] = value
                elif name in self.ref_features:
                    self.ref_features[name] = value
                    # If brand or model or variant was changed
                    # update product and related items, too
                    if self.ref_product and name in ("brand", "model", "variant"):
                        self._rename_product(self.ref_product, name, value)
                else:
                    self._add_to_ref(name, value)
                if product_to_add:
                    self.endResetModel()
                    self.emergency_resize.emit()
            return ok
        return False

    def extreme_validation(self, name: str, value: [str, int, float]) -> [bool, [str, int, float]]:
        feature_type = self._get_feature_type(name)
        if isinstance(value, str):
            value = value.strip()
        if feature_type == "e":
            value = str(value).lower()
            if value not in self.default_features["values"][name]:
                return False, None
        elif feature_type == "d":
            value = self._printable_to_value(name, value)
            value = float(value)
            if value <= 0:
                return False, None
        elif feature_type == "i":
            value = self._printable_to_value(name, value)
            value = int(round(value))
            if value <= 0:
                return False, None
        else:
            if len(value) <= 0:
                return False, None
        return True, value

    def _rename_product(self, product: dict, feature: str, value: str):
        if "brand" not in product or "model" not in product or "variant" not in product:
            # Sanity check, but should never happen
            return

        for maybe in self._data:
            if maybe.get("type") == "I":
                if ToolBoxWidget.bmv_match(product, maybe.get("features", {})):
                    maybe["features"][feature] = value

        # Finally, update product itself
        self.ref_product[feature] = value

    def removeRow(self, row: int, parent: QtCore.QModelIndex() = ...) -> bool:
        # TODO: remove this limitation asd
        if not self._pre_delete_check(row):
            return False

        self.beginRemoveRows(parent, row, row)
        try:
            removed = self.feature_keys.pop(row)
            try:
                del self.features[removed]
                # TODO: remove from features but show product one if it exists
                if removed in self.ref_features:
                    del self.ref_features[removed]
                if self.ref_product and removed in self.ref_product["features"]:
                    del self.ref_product["features"][removed]
            except IndexError:
                pass
        except IndexError:
            self.endRemoveRows()
            return False
        self.endRemoveRows()
        return True

    def _pre_delete_check(self, row) -> bool:
        feature_name = self.feature_keys[row]
        if self.features.get(feature_name) in ("brand", "model", "variant"):
            if self.ref_product:
                return False
        return True

    @staticmethod
    def _printable_to_value(name, value):
        # noinspection PyBroadException
        try:
            value = prettyprinter.printable_to_value(prettyprinter.name_to_unit(name), value)
        except BaseException:
            value = 0
        return value

    def _get_feature_type(self, name):
        feature_type = self.default_features["types"].get(name, "s")
        return feature_type

    def insert_row(self, feature: str, value: str) -> bool:
        if feature in self.feature_keys:
            return False

        row_index = self.rowCount()

        ok, value = self.extreme_validation(feature, value)
        product_to_add = None
        if ok:
            self.beginInsertRows(QtCore.QModelIndex(), row_index, row_index)
            self.feature_keys.append(feature)
            self.features[feature] = value
            if not self.ref_product:
                product_to_add = ToolBoxWidget.find_matching_product(self._data, self.features)
            self._add_to_ref(feature, value)
            self.endInsertRows()

        if product_to_add:
            self.beginResetModel()
            self._productize(self.features, product_to_add)
            self.endResetModel()

        self.emergency_resize.emit()

        return ok

    def features_in_table(self):
        return self.feature_keys

    def _add_to_ref(self, name: str, value):
        if name in item_only_features():
            target = self.ref_features
        elif self.ref_product:
            target = self.ref_product["features"]
        else:
            target = self.ref_features
        target[name] = value


class ToolBoxItem(QtWidgets.QWidget):
    def __init__(self, data: list[dict], features: dict, product: Optional[dict], default_features: dict):
        super().__init__()
        self.default_features = default_features

        self.main_layout = QtWidgets.QVBoxLayout()
        self.table = CustomTableView()
        self.features_combo_box = QtWidgets.QComboBox()
        self.feature_line_edit = QtWidgets.QLineEdit()
        self.feature_selector = QtWidgets.QComboBox()
        self.add_feature_button = QtWidgets.QPushButton("Add")
        self.add_feature_button.clicked.connect(self.add_feature)
        self.add_feature_button.setMinimumWidth(60)

        # setup
        self.table_setup(data, features, product, default_features)
        self.adder_layout = self._create_feature_adder()
        self.main_layout.addLayout(self.adder_layout)
        self.setLayout(self.main_layout)

    def external_size_hint_height(self):
        h1 = max(self.table.minimumSizeHint().height(), self.table.sizeHint().height())
        h2 = self.adder_layout.sizeHint().height()

        return h1 + h2

    def table_setup(self, data: list[dict], features: dict, product: Optional[dict], default_features: dict):
        ctm = CustomTableModel(data, features, product, default_features)
        ctm.emergency_resize.connect(self._do_the_emergency_resize)
        self.table.verticalHeader().hide()
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.setModel(ctm)
        self.table.setShowGrid(False)
        self.table.setAlternatingRowColors(True)
        self.table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.table.verticalScrollBar().hide()
        self.table.verticalScrollBar().resize(0, 0)
        self.main_layout.addWidget(self.table)
        self.table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.table.resizeColumnsToContents()
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)

        hh = QtWidgets.QHeaderView = self.table.horizontalHeader()
        hh.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        hh.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Stretch)

    def _do_the_emergency_resize(self):
        self.parentWidget().parentWidget().updateGeometry()
        self.table.resizeColumnToContents(0)

    def _create_feature_adder(self):
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.features_combo_box)
        layout.addWidget(self.feature_line_edit)
        layout.addWidget(self.feature_selector)
        layout.addWidget(self.add_feature_button)
        for name in self.default_features["names"]:
            self.features_combo_box.addItem(self.default_features["names"][name])
        self.features_combo_box.currentTextChanged.connect(self.set_input_type)
        self.set_input_type()
        return layout

    def set_input_type(self):
        for the_type in self.default_features["types"]:
            if self.default_features["names"][the_type] == self.features_combo_box.currentText():
                # set input type
                if self.default_features["types"][the_type] == "e":
                    self.feature_line_edit.setHidden(True)
                    self.feature_selector.setHidden(False)
                    self.feature_selector.clear()
                    i = 0
                    for value in self.default_features["values"][the_type]:
                        self.feature_selector.addItem(self.default_features["values"][the_type][value])
                        self.feature_selector.setItemData(i, value)
                        i += 1
                else:
                    self.feature_line_edit.setHidden(False)
                    self.feature_selector.setHidden(True)
                # disable button if feature already in table
                if the_type in self.table.model().features_in_table():
                    self.set_add_control_enabled(False)
                else:
                    self.set_add_control_enabled(True)

    def add_feature(self):
        feature = list(self.default_features["types"])[self.features_combo_box.currentIndex()]
        if self.feature_line_edit.isVisible():
            value = self.feature_line_edit.text()
        elif self.feature_selector.isVisible():
            value = self.feature_selector.currentData()
        else:
            return
        model: CustomTableModel = self.table.model()
        ok = model.insert_row(feature, value)
        if ok:
            self.set_add_control_enabled(False)
        self.feature_line_edit.clear()

    def set_add_control_enabled(self, flag: bool):
        self.add_feature_button.setEnabled(flag)
        self.feature_line_edit.setEnabled(flag)
        self.feature_selector.setEnabled(flag)

    def minimumSizeHint(self) -> QtCore.QSize:
        return self.table.minimumSizeHint()


class ToolBoxWidget(QtWidgets.QToolBox):
    def __init__(self, data: list[dict], default_features: dict, encountered_types_count: dict):
        super().__init__()
        self.data = data
        self.default_features = default_features
        self.encountered_types_count = encountered_types_count
        self.menu = None

        # variables
        self.encountered_types_current_count = defaultdict(lambda: 0)

    def clear(self):
        for idx in range(self.count()):
            self.removeItem(0)
        self.encountered_types_count.clear()
        self.encountered_types_current_count.clear()

    def load_items(self, data: list[dict]):
        if data:
            self.clear()
            self.data = data
            self.types_count()

        # find brand, model and variant of all products in data
        products = {}
        for idx, entry in enumerate(self.data):
            if entry["type"] == "P":
                products[idx] = (entry["brand"], entry["model"], entry["variant"])

        for entry in self.data:
            self.add_item(entry)

        # remove scroll in toolbox's scrollAreas
        for scroll_area in self.findChildren(QtWidgets.QScrollArea):
            scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            scroll_area.verticalScrollBar().hide()
            scroll_area.verticalScrollBar().resize(0, 0)

    def add_item(self, item: Optional[dict] = None, item_type: Optional[str] = None, single_item: Optional[bool] = False):
        if single_item:
            self.data.append(
                {
                    "type": "I",
                    "features": {
                        "type": item_type,
                    },
                }
            )
            found_product = None
        else:
            found_product = self.find_matching_product(self.data, item.get("features", {}))

        if item and item["type"] != "I":
            return

        counter = ""

        if item_type is None and item:
            item_type = item["features"]["type"]
        if self.encountered_types_count[item_type] > 1:
            self.encountered_types_current_count[item_type] += 1
            counter = f" #{self.encountered_types_current_count[item_type]}"
        if single_item:
            features = self.data[-1]["features"]
        else:
            features = item["features"]

        self.addItem(
            ToolBoxItem(self.data, features, found_product, self.default_features),
            f"{self.print_type_cool(item_type)}{counter}",
        )
        if item_type in ICON:
            icon = QtGui.QIcon(ICON[item_type])
            self.setItemIcon(self.count() - 1, icon)

        self.set_context_menu()

    @staticmethod
    def find_matching_product(data: list[dict], features: dict):
        if "model" in features and "brand" in features and "variant" in features:
            for maybe in data:
                if maybe.get("type") == "P":
                    if ToolBoxWidget.bmv_match(features, maybe):
                        return maybe
        return None

    @staticmethod
    def bmv_match(dict1, dict2) -> bool:
        return dict1.get("brand") == dict2.get("brand") and dict1.get("model") == dict2.get("model") and dict1.get("variant") == dict2.get("variant")

    def set_context_menu(self):
        counter = 0
        for item in self.children():
            if type(item) == QtWidgets.QAbstractButton:
                item: QtWidgets.QAbstractButton
                item.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
                item.customContextMenuRequested.connect(self.show_menu)
                counter += 1

    def print_type_cool(self, the_type: str) -> str:
        if the_type in self.default_features["values"]["type"]:
            return self.default_features["values"]["type"][the_type]
        else:
            return the_type.title()

    def show_menu(self):
        button = self.sender()
        self.menu = QtWidgets.QMenu()
        remove_action = QtGui.QAction("Remove item", self)
        remove_action.triggered.connect(lambda: self.remove_item_from_toolbox(button))
        self.menu.addAction(remove_action)
        self.menu.popup(QtGui.QCursor.pos())

    def minimumSizeHint(self) -> QtCore.QSize:
        h = 0
        for child in self.children():
            if isinstance(child, QtWidgets.QScrollArea):
                if child.isHidden():
                    # print("Hidden!")
                    pass
                    # print(f"Hidden min {child.minimumSizeHint().height()}")
                    # print(f"Hidden {child.sizeHint().height()}")
                    # h += child.minimumSizeHint().height()
                else:
                    the_widget = child.widget()
                    if the_widget and isinstance(the_widget, ToolBoxItem):
                        hinted = the_widget.external_size_hint_height()
                        # print(f"Hinted: {hinted}")
                        h += hinted
                    else:
                        # h += max(child.sizeHint().height(), child.minimumSizeHint().height())
                        pass
            elif isinstance(child, QtWidgets.QAbstractButton):
                # print(f"{child}: {child.sizeHint().height()} {child.minimumSizeHint().height()}")
                # Why 1.5? Dunno, they're ~40 pixels and sizeHint is 25 (minimum 24).
                h += int(child.sizeHint().height() * 1.5)
        old = super().minimumSizeHint()
        if h > old.height():
            return QtCore.QSize(old.width(), h)
        return old

    def remove_item_from_toolbox(self, button):
        i = 0
        for item in self.children():
            if type(item) == QtWidgets.QAbstractButton:
                if item == button:
                    self.removeItem(i)
                    break
                else:
                    i += 1

    def removeItem(self, index: int) -> None:
        i = 0
        data_index = None
        for data_index, entry in enumerate(self.data):
            if entry["type"] != "I":
                continue
            if index == i:
                break
            i += 1

        item_to_remove = self.data[data_index]
        item_b = item_to_remove["features"].get("brand")
        item_m = item_to_remove["features"].get("model")
        item_v = item_to_remove["features"].get("variant")
        counter = 0
        product_index = None
        deleted = False
        if item_b and item_m and item_v:
            for idx, entry in enumerate(self.data):
                # count items with the same product
                if entry["type"] == "I" and idx != data_index:
                    test_b = entry["features"].get("brand")
                    test_m = entry["features"].get("model")
                    test_v = entry["features"].get("variant")
                    if item_b == test_b and item_m == test_m and item_v == test_v:
                        counter += 1
                # find the product itself
                elif entry["type"] == "P":
                    p_test_b = entry.get("brand")
                    p_test_m = entry.get("model")
                    p_test_v = entry.get("variant")
                    if item_b == p_test_b and item_m == p_test_m and item_v == p_test_v:
                        product_index = idx
            if counter <= 0 and product_index:
                # If both item and product have to be deleted, delete them
                # without f...messing up indexes
                if data_index >= product_index:
                    del self.data[data_index]
                    del self.data[product_index]
                else:
                    del self.data[product_index]
                    del self.data[data_index]
                deleted = True

        # All other cases (item with no product, product not found, other items linked to product):
        # just delete the product
        if not deleted:
            del self.data[data_index]

        widget_ref = self.widget(index)
        super().removeItem(index)
        widget_ref.deleteLater()

    def types_count(self, data: list = None):
        if data is not None:
            self.data = data
        for entry in self.data:
            if entry["type"] != "I":
                continue
            the_type = entry["features"]["type"]
            self.encountered_types_count[the_type] += 1



import os
from pathlib import Path

from dotenv import load_dotenv

from .constants import basedir
from .commons import env_to_bool

home_dir = Path().home()
conf_dirs = [
    home_dir.joinpath(".config/peracotta"),
    home_dir.joinpath("peracotta"),
    Path("/etc/peracotta"),
    basedir,
]

for c_dir in conf_dirs:
    if c_dir.exists():
        conf_dir = c_dir
        break


def load_conf():
    load_dotenv(conf_dir.joinpath(".env"))


load_conf()

CONFIG = {
    "TARALLO_URL": os.environ.get("TARALLO_URL", None),
    "TARALLO_TOKEN": os.environ.get("TARALLO_TOKEN", None),
    "TARALLO_FEATURES_AUTO_DOWNLOAD": env_to_bool(os.environ.get("TARALLO_FEATURES_AUTO_DOWNLOAD", "1")),
    "GENERATE_FILES_USE_SUDO": env_to_bool(os.environ.get("GENERATE_FILES_USE_SUDO", "1")),
    "GENERATE_FILES_ASK_SUDO_PASSWORD": env_to_bool(os.environ.get("GENERATE_FILES_ASK_SUDO_PASSWORD", "1")),
}
